"""
    ISubfinder.py

    Nokutoka Momiji

    Contains the ISubfinder interface to be used for creating Subfinders, as well as the SubfinderConfiguration class
    for configuring Subfinder properties.

    This ensures that all Subfinders must provide the same methods, returning data in the exact same manner, bringing uniformity.
"""

from abc import ABC, abstractmethod
from typing import Callable, Any

from nokufind import Post, Comment, Note
from nokufind.Utils import PIXIV_REFERER

class SubfinderConfiguration():
    """A class for configuring Subfinder properties"""

    def __init__(self, callback: Callable[[str, Any, bool, bool], None] = None):
        """Initializes the configuration object.

        Args:
            callback (``Callable[[str, Any, bool, bool], None]``, optional): Callback function to execute when a change to the settings is made. Defaults to None.
        """
        self.__config = {}
        self.__config["cookies"] = {
            "cf_clearance": ""
        }
        self.__config["headers"] = {
            "User-Agent": "",
            "Referer": PIXIV_REFERER
        }
        self.__config["api_key"] = None
        self.__callback = callback

    def set_cookie(self, key: str, value: str | bytes) -> None:
        """Sets a request cookie.

        Args:
            key (str): The name of the cookie
            value (str | bytes): The value of the cookie
        """
        self._check_valid_value_type(key)
        self._check_valid_value_type(value)

        self.__config["cookies"][key] = value
        
        if (callable(self.__callback)):
            self.__callback(key, value, True, False)

    def get_cookie(self, key: str) -> str | bytes:
        """Returns a request cookie

        Args:
            key (str): The name of the cookie.

        Returns:
            ``str | bytes``: The value of the cookie. 
        """
        self._check_valid_value_type(key)
        
        return self.__config["cookies"][key]
    
    def set_header(self, key: str, value: str | bytes):
        """Sets a request header key.

        Args:
            key (``str``): The name of the header key.
            value (``str | bytes``): The value of the header key.
        """
        self._check_valid_value_type(key)
        self._check_valid_value_type(value)

        self.__config["headers"][key] = value
        
        if (callable(self.__callback)):
            self.__callback(key, value, False, True)

    def get_header(self, key: str) -> str | bytes:
        """Returns a request header value

        Args:
            key (``str``): The key of the request header value.

        Returns:
            ``str | bytes``: The request header key value.
        """
        self._check_valid_value_type(key)

        return self.__config["headers"][key]
    
    def set_config(self, key: str, value):
        """Sets the value for a setting.

        Args:
            key (``str``): The name of the setting.
            value (``Any``): The value for the setting.

        Raises:
            ``ValueError``: Raised if the key is a header key or a cookie.
        """
        if (key in ["cookies", "headers"] or key not in self.__config):
            raise ValueError(f"Invalid key \"{key}\".")
        
        self.__config[key] = value

        if (callable(self.__callback)):
            self.__callback(key, value, False, False)

    def get_config(self, key: str, default_value = None):
        """Returns the value of a setting. 
        
        If the setting is not found, it returns the default value.

        Args:
            key (``str``): The name of the setting.
            default_value (``Any``, optional): The default value to return if the setting is not found. Defaults to None.

        Returns:
            ``Any``: The value of the setting, or the default value if the setting does not exist.
        """
        if (key == "cookies"):
            return self.cookies
        
        if (key == "headers"):
            return self.headers
        
        return self.__config.get(key, default_value)
    
    def _set_property(self, key: str, default_value):
        """(Internal) This function is only meant for Subfinders to use for adding the settings they require for functioning.

        ATTENTION! Do not use this function, use ``set_header``, ``set_cookie`` or ``set_config`` instead.

        Args:
            key (``str``): The name of the setting.
            default_value (``Any``): The default value for the setting.

        Raises:
            ``ValueError``: Raised if an attempt is made to configure header values or cookies.
        """
        if (key in ["cookies", "headers"]):
            raise ValueError("Do not use _set_property for configuring headers or cookies, use the respective functions.")
        
        self.__config[key] = default_value

    def _check_valid_value_type(self, value):
        """(Internal) Used to check that the value's type is valid for request header / cookie use.

        Args:
            value (Any): Value to check. 

        Raises:
            ``TypeError``: Raised if the value is not of type ``str`` or ``bytes``
        """
        if (type(value) not in [str, bytes]):
            raise TypeError(f"Value should be of type str or bytes, received {type(value)}.")
    
    def __getitem__(self, key: str):
        return self.get_config(key)

    @property
    def headers(self) -> dict[str | bytes]:
        """Returns the headers as configured.

        Returns:
            ``dict[str | bytes]``: A dictionary containing the request headers as configured.
        """
        return self.__config["headers"].copy()
    
    @property
    def cookies(self) -> dict[str | bytes]:
        """Returns the cookies as configured.

        Returns:
            ``dict[str | bytes]``: A dictionary containing the request cookies as configured.
        """
        return self.__config["cookies"].copy()

    @property
    def cf_clearance(self) -> str:
        """Used for returning the "cf_clearance" cookie value."""
        return self.__config["cookies"]["cf_clearance"]
    
    @property
    def user_agent(self) -> str:
        """Returns the request "User-Agent" header value."""
        return self.__config["headers"]["User-Agent"]
    
    @property
    def referer(self) -> str:
        """Returns the request "Referer" header value"""
        return self.__config["headers"]["Referer"]
    
    @property
    def api_key(self) -> str:
        """Returns an API key, if set."""
        return self.__config["api_key"]

class ISubfinder(ABC):
    # Static methods.

    """
        All Finders must implement functions to transform the data acquired from the API
        into a general Post, Comment or Note object.
    """

    @staticmethod
    @abstractmethod
    def to_post(post_data) -> Post:
        """(Internal) Creates a Post object from the given data.

        Args:
            post_data (``Any``): An object containing all of the data from the post, acquired from the API.

        Returns:
            ``Post``: A Post object containing the extracted data.
        """
        ...

    @staticmethod
    @abstractmethod
    def to_comment(comment_data) -> Comment:
        """(Internal) Creates a Comment object from the given data.

        Args:
            comment_data (``Any``): An object containing all of the data from the comment, acquired from the API.

        Returns:
            ``Comment``: A Comment object containing the extracted data.
        """
        ... 

    @staticmethod
    @abstractmethod
    def to_note(note_data) -> Note:
        """(Internal) Creates a Note object from the given data.

        Args:
            note_data (``Any``): An object containing all of the data from the note, acquired from the API.

        Returns:
            ``Note``: A Note object containing the extracted data.
        """
        ...

    @abstractmethod
    def search_posts(self, tags: str | list[str], *, limit: int = 100, page: int | None = None) -> list[Post]:
        """Searches for posts that match the given tags.

        Args:
            tags (``str | list[str]``): A string or list of tags to match against.
            limit (``int``, optional): The maximum number of posts to get. Defaults to 100.
            page (``int | None``, optional): The page number. Defaults to None.

        Returns:
            ``list[Post]``: A list containing all the found posts.
        """
        ...

    @abstractmethod
    def get_post(self, post_id: int) -> Post | None:
        """Searches for a post with the given ID.

        Args:
            post_id (``int``): The ID of the post to look for.

        Returns:
            ``Post | None``: The post, or None if a post with the ID was not found.
        """
        ...

    @abstractmethod
    def search_comments(self, *, post_id: int = None, limit: int | None = None, page: int | None = None) -> list[Comment]:
        """Searches for comments.

        Args:
            post_id (``int``, optional): The ID of the post to get the comments from. Defaults to None.
            limit (``int | None``, optional): The maximum number of comments to get. Defaults to None.
            page (``int | None``, optional): The page number. Defaults to None.

        Returns:
            ``list[Comment]``: A list containing the found comments.
        """
        ...

    @abstractmethod
    def get_comment(self, comment_id: int) -> Comment | None:
        """Searches for a comment with the given ID.

        Args:
            comment_id (``int``): The ID of the comment to look for.

        Returns:
            ``Comment | None``: The comment, or None if a matching comment was not found.
        """
        ...

    @abstractmethod
    def get_notes(self, post_id: int) -> list[Note]:
        """Searches for the notes in a given post.

        Args:
            post_id (``int``): The ID of the post to get the notes from.

        Returns:
            ``list[Note]``: A list containing the found notes.
        """
        ...

    @abstractmethod
    def post_get_parent(self, post: Post) -> Post | None:
        """Searches for the parent post of a post.

        Args:
            post (``Post``): The post.

        Returns:
            ``Post | None``: The parent post, or None if the post has no parent.
        """
        ...

    @abstractmethod
    def post_get_children(self, post: Post) -> list[Post]:
        """Searches for the children posts of a post.

        Args:
            post (``Post``): The post.

        Returns:
            ``list[Post]``: A list containing the children of the post.
        """
        ...

    @property
    @abstractmethod
    def configuration(self) -> SubfinderConfiguration:
        """The Subfinder configuration object.

        This can be used to set any settings related to the current Subfinder.

        Returns:
            SubfinderConfiguration: The Subfinder configuration object.
        """
        ...


"""
Nokufind
~~~~~~~~

A library that allows you to find posts from multiple Boorus and sources.
It also allows you to create custom wrappers around other finders and API modules.
"""

__title__ = "nokufind"
__author__ = "Nokutoka Momiji"
__license__ = "MIT"
__copyright__ = "Copyright (C) 2024 Nokutoka Momiji"
__version__ = "1.0.0"

__path__ = __import__("pkgutil").extend_path(__path__, __name__)

from nokufind.Post import Post
from nokufind.Comment import Comment
from nokufind.Note import Note
from nokufind.Finder import Finder
from nokufind.Subfinder import builtin_finders
from nokufind.Utils import Utils


import requests
from datetime import datetime
from calendar import timegm
from html.parser import HTMLParser
from time import sleep

from pybooru import Moebooru, PybooruHTTPError
from nokufind import Comment, Note, Post

from nokufind.Subfinder import ISubfinder, SubfinderConfiguration
from nokufind.Utils import log, make_request, assert_conversion 


class KonaParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.comment_ids = []
        self.in_popular_list = False
        self.popular_posts = []

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
        if tag == "div" and ("class", "comment avatar-container") in attrs:
            comment_id = int(attrs[1][1][1:])
            self.comment_ids.append(comment_id)

        if tag == "ul" and ("id", "post-list-posts") in attrs:
            self.in_popular_list = True

        if tag == "li" and self.in_popular_list:
            try:
                possible_id = attrs[1][1]
                if not possible_id.startswith("p"):
                    return
                self.popular_posts.append(int(possible_id[1:]))
            except Exception as e:
                print(e)
    def handle_endtag(self, tag: str) -> None:
        if tag == "ul" and self.in_popular_list:
            self.in_popular_list = False

    def find_comments(self, html):
        self.comment_ids = []
        self.feed(html)
        return self.comment_ids
    
    def find_popular_posts(self, html):
        self.popular_posts = []
        self.feed(html)
        return self.popular_posts
        

class KonachanFinder(ISubfinder):
    @staticmethod
    def to_post(post_data) -> Post:
        return Post(
            post_id = post_data["id"],
            tags = post_data["tags"].split(" "),
            sources = post_data["source"].split(" "),
            images = [post_data["file_url"]],
            authors = [""],
            source = "konachan",
            preview = post_data["preview_url"],
            md5 = [post_data["md5"]],
            rating = post_data["rating"],
            parent_id = post_data["parent_id"],
            dimensions = [(post_data["width"], post_data["height"])],
            poster = post_data["author"],
            poster_id = post_data["creator_id"],
            name = None
        )
    
    @staticmethod
    def to_comment(comment_data) -> Comment:
        return Comment(
            comment_id = comment_data["id"],
            post_id = comment_data["post_id"],
            creator_id = comment_data["creator_id"],
            creator = comment_data["creator"],
            body = comment_data["body"],
            source = "konachan",
            created_at = datetime.strptime(comment_data["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ")
        )

    @staticmethod
    def to_note(note_data) -> Note:
        return Note(
            note_id = note_data["id"],
            created_at = datetime.strptime(note_data["created_at"], "%Y-%m-%dT%H:%M:%S.%fZ"),
            x = note_data["x"],
            y = note_data["y"],
            width = note_data["width"],
            height = note_data["height"],
            body = note_data["body"],
            source = "konachan",
            post_id = note_data["body"]
        )
    
    def __init__(self, username = "", password = "", hash_string = ""):
        self.__client_name = "konachan"

        self.__client = Moebooru(self.__client_name, username = username, password = password, hash_string = hash_string)
        self.__parser = KonaParser()
        
        self.__config = SubfinderConfiguration()
        self.__config._set_property("username", username)
        self.__config._set_property("password", password)
        self.__config._set_property("hash_string", hash_string)

        self.__name = f"nokufind.Subfinder.{self.__class__.__name__}"
        log(f"> [{self.__name}]: Initialized a new {self.__class__.__name__} instance.")

    def search_posts(self, tags: str | list[str], *, limit: int = 100, page: int | None = None) -> list[Post]:
        self._check_client()
        
        tags = " ".join(tags) if (type(tags) == list) else tags
        page = page if page else 1
        raw_posts = self._get_all_posts(tags, limit, page)

        return [KonachanFinder.to_post(post) for post in raw_posts]
    
    def get_post(self, post_id: int) -> Post | None:
        self._check_client()

        post_id = assert_conversion(post_id, int, "post_id")

        try:
            post = self.search_posts(f"id:{post_id}")
        except PybooruHTTPError as e:
            sleep(2)
            return self.get_post(post_id)
        except Exception as e:
            log(f"> [{self.__name}]: Failed to get post id {post_id}.\nException: {e}")

        return post[0] if post else None
    
    def search_comments(self, *, post_id=None, limit=None, page=None) -> list[Comment]:
        self._check_client()

        limit = 100 if limit == None else limit
        page = 1 if page == None else page

        return self._get_comments(post_id, limit, page)

    def get_comment(self, comment_id: int) -> Comment | None:
        self._check_client()

        comment_id = assert_conversion(comment_id, int, "comment_id")
        try:
            raw_comment = self.__client.comment_show(comment_id)
            return KonachanFinder.to_comment(raw_comment)
        except Exception as e:
            log(f"> [{self.__name}]: Failed to get comment id {comment_id}.\n{e}")
            return None
    
    def get_notes(self, post_id: int) -> list[Note]:
        self._check_client()

        post_id = assert_conversion(post_id)

        try:
            raw_notes = self.__client.note_list(post_id = post_id)
            return [KonachanFinder.to_note(note) for note in raw_notes]
        except Exception as e:
            log(f"> [{self.__name}]: Failed to get notes from post {post_id}.\nException: {e}")
            return []
        
    def post_get_parent(self, post: Post) -> Post | None:
        self._check_client()

        if (type(post) != Post):
            raise TypeError(f"\"post\" should be of type Post, not {post}.")
        
        if (not post.parent_id):
            return None

        parent_post = self.get_post(post.parent_id)

        return post._set_parent(parent_post)
    
    def post_get_children(self, post: Post) -> list[Post]:
        self._check_client()

        if (type(post) != Post):
            raise TypeError(f"\"post\" should be of type Post, not {type(post)}")
        
        children_posts = self.search_posts(f"parent:{post.post_id}")

        for index, child_post in enumerate(children_posts):
            if (child_post.post_id == post.post_id):
                children_posts.pop(index)
                break

        return post._set_children(children_posts)

    # Non-standard methods.
    def get_popular_posts(self, *, date: datetime | None = None, is_month = False) -> list[Post]:
        """
        Returns a list containing the most popular posts.

        By default, the function returns the most popular posts in the last 24 hours.
        However, if you set a ``date``, it will return the most popular posts of the week of the date.

        Alternatively, you can set ``is_month`` to ``True`` to get the popular posts of the month.

        Args:
            date (datetime, optional): The date for the week to check. If ``is_month`` is set to True, it will return the posts of the month instead. Defaults to None.
            is_month (bool, optional): Whether the function should return the posts of the month or the week. Defaults to False.

        Returns:
            list[Post]: A list of the most popular posts of the given timeframe (Default: last 24 hours).
        """
        self._check_client()

        url = f"{self.__client.site_url}/post/popular_recent"

        if (type(date) == datetime):
            url = f"{self.__client.site_url}/post/"
            if is_month:
                url += f"popular_by_month?month={date.month}&year={date.year}"
            else:
                url += f"popular_by_week?day={date.day - date.weekday()}&month={date.month}&year={date.year}"

        request = self._request(url)

        if (request.status_code >= 400):
            log(f"> [{self.__name}]: Url \"{url}\" returned error status code {request.status_code}\n{request.text}")
            return []
        
        post_ids = self.__parser.find_popular_posts(request.text)
        raw_posts = [self.get_post(post) for post in post_ids]

        posts = []

        for post in raw_posts:
            if (not post):
                continue
            posts.append(post)
            
        return posts

    def on_config_change(self, key: str, value, is_cookie: bool, is_header: bool):
        if key in ["username", "password", "hash_string"]:
            username = self.__config.get_config("username")
            password = self.__config.get_config("password")
            hash_string = self.__config.get_config("hash_string")
            self.__client = Moebooru(self.__client_name, username, password, hash_string)

    def _get_all_posts(self, tags: str, limit: int, page: int | None) -> list[dict]:
        all_posts = []
        current_page = page if page != None else 0
        previous_size = 100
        oversized = False


        while (previous_size == 100):
            try:
                raw_posts = self.__client.post_list(tags = tags, limit = 100, page = page)
            except KeyError:
                break
            except PybooruHTTPError as e:
                log(f"> [{self.__name}]: PybooruHTTPError: {e}")
                if "410" in e._msg:
                    break
                else:
                    sleep(0.1)
                    continue
            previous_size = len(raw_posts)

            all_posts += raw_posts

            if len(all_posts) >= limit:
                oversized = True
                break

            current_page += 1

        if oversized:
            return all_posts[:limit]
        
        return all_posts

    def _get_comments_of_post(self, post_id: int):
        log("In _get_comments_of_post")
        post_id = assert_conversion(post_id, int, "post_id")
        url = f"{self.__client.site_url}/post/show/{post_id}"

        request = requests.get(url)

        if (request.status_code >= 400):
            log(f"> [{self.__name}]: Url \"{url}\" returned error status code {request.status_code}.")
            return []
        
        comment_ids = self.__parser.find_comments(request.text)

        if not comment_ids:
            return []
        
        raw_comments = [self.get_comment(comment) for comment in comment_ids]

        comments = []
        for comment in raw_comments:
            if not comment:
                continue
            comments.append(comment)

        return comments


    def _get_comments(self, post_id: int | None, limit: int = 100, page: int = 1):
        if post_id != None:
            return self._get_comments_of_post(post_id)
        
        url = f"{self.__client.site_url}/comment"

        current_page = page
        all_comments = []
        oversized = False

        while True:
            request_url = url + f"?page={current_page}"
            request = requests.get(request_url)

            if (request.status_code >= 400):
                log(f"> [{self.__name}]: Url \"{request_url}\" returned status code {request.status_code}.")
                break

            comment_ids = self.__parser.find_comments(request.text)
            all_comments += comment_ids

            if (len(all_comments) >= limit):
                oversized = True
                break
        
        all_raw_comments = list(set(all_comments))
        all_comments.clear()
        for comment in all_raw_comments:
            if comment == None:
                continue
            all_comments.append(self.get_comment(comment))

        if oversized:
            return all_comments[:limit]
        
        return all_comments

    def _request(self, url, post = False):
        return make_request(url, post = post, headers = self.configuration.headers, cookies = self.configuration.cookies)

    def _check_client(self):
        if (not isinstance(self.__client, Moebooru)):
            raise RuntimeError(f"Client must be of type Moebooru, not {type(self.__client)}.")
        
    @property
    def configuration(self) -> SubfinderConfiguration:
        return self.__config


import asyncio
import xml.etree.ElementTree as ET
from datetime import datetime
from calendar import timegm
from itertools import combinations
from html.parser import HTMLParser

from pygelbooru import Gelbooru
from pygelbooru.gelbooru import GelbooruImage, GelbooruComment

from nokufind import Post, Comment, Note
from nokufind.Subfinder import ISubfinder, SubfinderConfiguration
from nokufind.Utils import log, assert_conversion, make_request, parse_tags, get

def _parse_notes_data(data):
    # Parse the XML-like data
    root = ET.fromstring(data)

    # Initialize a list to store formatted notes
    formatted_notes = []

    # Loop through each 'note' element
    for note_elem in root.findall('note'):
        note_data = {
            'id': note_elem.get('id'),
            'created_at': note_elem.get('created_at'),
            'updated_at': note_elem.get('updated_at'),
            'body': note_elem.get('body'),
            'position': {'x': note_elem.get('x'), 'y': note_elem.get('y')},
            'dimensions': {'width': note_elem.get('width'), 'height': note_elem.get('height')},
            'creator_id': note_elem.get('creator_id'),
        }

        formatted_notes.append(note_data)

    return formatted_notes

def _post_is_valid(post: GelbooruImage) -> bool:
    return post != None

def _filter_invalid_posts(posts: list[GelbooruImage]) -> list[dict]:
    return [post for post in posts if _post_is_valid(post)]

class GelbooruParser(HTMLParser):
    def __init__(self, *, convert_charrefs: bool = True) -> None:
        super().__init__(convert_charrefs=convert_charrefs)

        self.in_comment_body = False
        self.in_link = False
        self.in_comment_thumb = False
        self.comment_data = []
        self.last_comment = {}
        self.gelbooru = None

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
        if tag == "div" and ("class", "commentBody ") in attrs:
            print(f"{attrs}")
            self.in_comment_body = True
            return
        elif tag == "div" and ("class", "commentThumbnail") in attrs:
            self.in_comment_thumb = True

        if tag == "a" and self.in_comment_body and not self.last_comment.get("@creator_id", None):
            print(f"{attrs}")
            self.last_comment["@creator_id"] = int(attrs[0][1].split("=")[-1])
            self.in_link = True
        elif tag == "a" and self.in_comment_thumb and not self.last_comment.get("@post_id", None):
            self.last_comment["@post_id"] = int(attrs[0][1].split("=")[-1])

    def handle_data(self, data: str) -> None:
        if self.in_comment_body and self.in_link:
            self.last_comment["@creator"] = data.replace("<b>", "")
            return
        
        if self.in_comment_body:
            content = data.split("\n")
            if " commented at " in content[0]:
                content = content[0].split(" commented at ")[-1].split(" » ")
                self.last_comment["@created_at"] = content[0][:-3]
                self.last_comment["@id"] = content[1][1:]
            elif self.last_comment.get("@id") and not self.last_comment.get("@body", None) and content[0] != '':
                self.last_comment["@body"] = "\n".join(content)

    def handle_endtag(self, tag: str) -> None:
        if tag == "a" and self.in_link:
            self.in_link = False

        if tag == "div" and self.in_comment_body:
            self.in_comment_body = False
            self.comment_data.append(GelbooruComment(self.last_comment, gelbooru = self.gelbooru))
            self.last_comment = {}
        elif tag == "div" and self.in_comment_thumb:
            self.in_comment_thumb = False

    def get_comments(self, html: str, gelbooru: Gelbooru):
        self.gelbooru = gelbooru
        self.comment_data = []
        self.last_comment = {}
        self.feed(html)
        return self.comment_data

class GelbooruFinder(ISubfinder):
    @staticmethod
    def to_post(post_data: GelbooruImage) -> Post:

        if post_data.source == None:
            post_data.source = ""

        if (not hasattr(post_data, "_payload")):
            print(post_data.__dict__)

        post_data.parent_id = int(post_data._payload["parent_id"])

        return Post(
            post_id = post_data.id,
            tags = post_data.tags,
            sources = post_data.source.split(" "),
            images = [post_data.file_url],
            authors = [""],
            source = "gelbooru",
            preview = post_data._payload["preview_url"],
            md5 = [post_data._payload["md5"]],
            rating = post_data.rating,
            parent_id = post_data.parent_id if post_data.parent_id != 0 else None,
            dimensions = [(post_data.width, post_data.height)],
            poster = post_data._payload["owner"],
            poster_id = int(post_data._payload["creator_id"]),
            name = None
        )
    
    @staticmethod
    def to_comment(comment_data: GelbooruComment) -> Comment:
        return Comment(
            comment_id = comment_data.id,
            post_id = comment_data.post_id,
            creator_id = comment_data.creator_id,
            creator = comment_data.creator,
            body = comment_data.body,
            source = "gelbooru",
            created_at = comment_data.created_at
        )
    
    @staticmethod
    def to_note(note_data: dict) -> Note:
        return Note(
            note_id = note_data["id"],
            created_at = datetime.strptime(note_data["created_at"], "%a %b %d %H:%M:%S %z %Y"),
            x = int(note_data["position"]["x"]),
            y = int(note_data["position"]["y"]),
            width = int(note_data["dimensions"]["width"]),
            height = int(note_data["dimensions"]["height"]),
            body = note_data["body"],
            source = "gelbooru",
            post_id = note_data["post_id"]
        )

    def __init__(self, api_key = None, user_id = None) -> None:
        self.__client = Gelbooru(api_key, user_id)
        self.__config = SubfinderConfiguration()
        self.__config._set_property("api_key", None)
        self.__config._set_property("user_id", None)

        self.__name = f"nokufind.Subfinder.{self.__class__.__name__}"
        log(f"> [{self.__name}]: Initialized a new {self.__class__.__name__} instance.")

    def search_posts(self, tags: str | list[str], *, limit: int = 100, page: int | None = None) -> list[Post]:
        self._check_client()

        if (type(tags) == str):
            tags = tags.split(" ")

        posts = self._get_all_posts(tags = tags, limit = limit, page = page)

        return posts

    def get_post(self, post_id: int) -> Post | None:
        self._check_client()

        post_id = assert_conversion(post_id, int, "post_id")

        post = asyncio.run(self.__client.get_post(post_id))

        return GelbooruFinder.to_post(post) if post else None
    
    def search_comments(self, *, post_id=None, limit=None, page=None) -> list[Comment]:
        self._check_client()

        if post_id != None:
            post_id = assert_conversion(post_id, int, "post_id")
            
            try:
                raw_comments = asyncio.run(self.__client.get_comments(post_id))
                return [GelbooruFinder.to_comment(comment) for comment in raw_comments]
            except Exception as e:
                log(f"> [{self.__name}]: Getting comments from post failed. This is probably due to Gelbooru having disabled their comment API due to abuse.\n{e}")
                return []
        
    def get_comment(self, comment_id: int) -> Comment | None:
        self._check_client()

        comment_id = assert_conversion(comment_id, int, "comment_id")

        return None

    def get_notes(self, post_id: int) -> list[Note]:
        self._check_client()

        post_id = assert_conversion(post_id, int, "post_id")

        request = make_request(f"https://gelbooru.com/index.php?page=dapi&s=note&q=index&post_id={post_id}&json=1")

        if (request.status_code >= 400):
            log(f"> [{self.__name}]: Failed to get notes for post {post_id} ({request.status_code}).\n{request.text}")
            return []
        
        raw_notes = _parse_notes_data(request.text)

        for raw_note in raw_notes:
            raw_note["post_id"] = post_id

        return [GelbooruFinder.to_note(note) for note in raw_notes]
    
    def post_get_parent(self, post: Post) -> Post | None:
        self._check_client()
        
        if (type(post) != Post):
            raise TypeError(f"\"post\" must be of type Post, not {type(post)}.")
        
        if (not post.parent_id):
            return None
        
        parent_post = self.get_post(post.parent_id)

        return post._set_parent(parent_post)
    
    def post_get_children(self, post: Post) -> list[Post]:
        self._check_client()

        if (type(post) != Post):
            raise TypeError(f"\"post\" must be of type Post, not {type(post)}.")
        
        children_posts = self.search_posts(f"parent:{post.post_id}")

        for index, child_post in enumerate(children_posts):
            if child_post.post_id == post.post_id:
                children_posts.pop(index)
                break

        return post._set_children(children_posts)
    
    def _get_all_posts(self, tags: str, limit: int = 100, page: int | None = None) -> list[Post]:
        current_posts = []
        current_size = 100
        current_page = page if type(page) == int else 1
        
        while (current_size == 100):
            raw_posts = asyncio.run(self.__client.search_posts(tags = tags, limit = 100, page = current_page))
            current_size = len(raw_posts)

            if (type(raw_posts) == list):
                raw_posts = _filter_invalid_posts(raw_posts)
                posts = [GelbooruFinder.to_post(post) for post in raw_posts]
            elif _post_is_valid(raw_posts):
                posts = [GelbooruFinder.to_post(raw_posts)]
                
            current_posts += posts

            if len(current_posts) >= limit:
                break

            current_page += 1

        current_posts = list(sorted(current_posts, key = lambda x: x.post_id))

        if len(current_posts) > limit:
            return current_posts[:limit]
        
        return current_posts

    def _check_client(self):
        if (not isinstance(self.__client, Gelbooru)):
            raise RuntimeError(f"Client should be of type Gelbooru, not {type(self.__client)}")

    @property
    def configuration(self):
        return self.__config


"""
    PixivAuth.py

    Nokutoka Momiji

    This script derives from ZipFile's Pixiv OAuth script (https://gist.github.com/ZipFile/c9ebedb224406f4f11845ab700124362)
"""

import requests
import platform
import io
import time
import json
import re
import os
from base64 import urlsafe_b64encode
from hashlib import sha256
from secrets import token_urlsafe
from urllib.parse import urlencode
from zipfile import ZipFile

from appdirs import user_data_dir
from selenium import webdriver
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities

from nokufind.Utils import log

CHROMEDRIVER_URL = "https://chromedriver.storage.googleapis.com/91.0.4472.101/"
CHROMEDRIVER_FILES = {
    "Windows": "chromedriver_win32.zip",
    "Linux": "chromedriver_linux64.zip",
    "Darwin": "chromedriver_mac64.zip",
    "DarwinM1": "chromedriver_mac64_m1.zip"
}

# Latest app version can be found using GET /v1/application-info/android
USER_AGENT = "PixivIOSApp/7.13.3 (iOS 14.6; iPhone13,2)"
REDIRECT_URI = "https://app-api.pixiv.net/web/v1/users/auth/pixiv/callback"
LOGIN_URL = "https://app-api.pixiv.net/web/v1/login"
AUTH_TOKEN_URL = "https://oauth.secure.pixiv.net/auth/token"
CLIENT_ID = "MOBrBDS8blbauoSck0ZfDbtuzpyT"
CLIENT_SECRET = "lsACyCD94FhDUtGTXi3QzcFE2uU1hqtDaKeqrdwj"
REQUESTS_KWARGS = {
    # 'proxies': {
    #     'https': 'http://127.0.0.1:1087',
    # },
    # 'verify': False
}

def install_chromedriver() -> bool:
    current_platform = platform.system()
    
    if (current_platform == "Darwin"):
        current_platform += "M1" if platform.processor() == "arm" else ""
    
    full_url = f"{CHROMEDRIVER_URL}{CHROMEDRIVER_FILES[current_platform]}"

    request = requests.get(full_url)

    if (request.status_code >= 400):
        return False
    
    save_dir = user_data_dir("chromedriver")

    with ZipFile(io.BytesIO(request.content), "r") as zip_file:
        zip_file.extractall(save_dir)

    return True
    
def s256(data) -> bytes:
    """S256 transformation method."""

    return urlsafe_b64encode(sha256(data).digest()).rstrip(b"=").decode("ascii")


def oauth_pkce(transform) -> tuple[str, ...]:
    """Proof Key for Code Exchange by OAuth Public Clients (RFC7636)."""

    code_verifier = token_urlsafe(32)
    code_challenge = transform(code_verifier.encode("ascii"))

    return code_verifier, code_challenge


def print_auth_token_response(response):
    data = response.json()

    try:
        access_token = data["access_token"]
        refresh_token = data["refresh_token"]
    except KeyError:
        log("error:")
        log(data)

    log(f"access_token: {access_token}")
    log(f"refresh_token: {refresh_token}")
    log(f"expires_in: {data.get('expires_in', 0)}")
    return refresh_token


def login():
    path = user_data_dir("chromedriver")
    
    if (not os.path.isdir(path)):
        log("> Chromedriver not found. Installing...")
        install_chromedriver()

    caps = DesiredCapabilities.CHROME.copy()
    caps["goog:loggingPrefs"] = {"performance": "ALL"}  # enable performance logs
    
    options = webdriver.ChromeOptions()
    options.set_capability("goog:loggingPrefs", {"performance": "ALL"})
    options.add_experimental_option("excludeSwitches", ["enable-automation"])
    options.add_experimental_option('useAutomationExtension', False)
    service = webdriver.ChromeService()
    driver = webdriver.Chrome(options = options, service = service)

    code_verifier, code_challenge = oauth_pkce(s256)
    login_params = {
        "code_challenge": code_challenge,
        "code_challenge_method": "S256",
        "client": "pixiv-android",
    }

    log(f"> Gen code_verifier: {code_verifier}")

    driver.get(f"{LOGIN_URL}?{urlencode(login_params)}")

    while True:
        # wait for login
        if driver.current_url[:40] == "https://accounts.pixiv.net/post-redirect":
            break
        time.sleep(1)

    # filter code url from performance logs
    code = None
    for row in driver.get_log('performance'):
        data = json.loads(row.get("message", {}))
        message = data.get("message", {})
        if message.get("method") == "Network.requestWillBeSent":
            url = message.get("params", {}).get("documentURL")
            if url[:8] == "pixiv://":
                code = re.search(r'code=([^&]*)', url).groups()[0]
                break

    driver.close()

    log(f"> Get code: {code}")

    response = requests.post(
        AUTH_TOKEN_URL,
        data={
            "client_id": CLIENT_ID,
            "client_secret": CLIENT_SECRET,
            "code": code,
            "code_verifier": code_verifier,
            "grant_type": "authorization_code",
            "include_policy": "true",
            "redirect_uri": REDIRECT_URI,
        },
        headers={
            "user-agent": USER_AGENT,
            "app-os-version": "14.6",
            "app-os": "ios",
        },
        **REQUESTS_KWARGS
    )

    return print_auth_token_response(response)


def refresh(refresh_token):
    response = requests.post(
        AUTH_TOKEN_URL,
        data={
            "client_id": CLIENT_ID,
            "client_secret": CLIENT_SECRET,
            "grant_type": "refresh_token",
            "include_policy": "true",
            "refresh_token": refresh_token,
        },
        headers={
            "user-agent": USER_AGENT,
            "app-os-version": "14.6",
            "app-os": "ios",
        },
        **REQUESTS_KWARGS
    )

    return print_auth_token_response(response)


"""
    Post.py

    Nokutoka Momiji


    Contains a Post class representing a wrapper around post data from multiple sources. 
    
    The Post class wraps information such as post ID, tags, sources, images, authors, source, preview, 
    MD5 hashes, rating, parent ID, dimensions, poster, and poster ID. 
    
    It also provides methods for downloading and fetching image data.

    This script relies on built-in Python modules such as os, hashlib, json, threading, and concurrent.futures, 
    as well as external libraries, such as tqdm for progress bars.

"""


from __future__ import annotations

import os
import hashlib
import json
import threading
from concurrent.futures import ThreadPoolExecutor
import requests
from enum import Enum
from reprlib import Repr

from tqdm import tqdm

from nokufind.Utils import make_request, log, expand

class Rating(int, Enum):
        GENERAL = 1
        SENSITIVE = 2
        QUESTIONABLE = 3
        EXPLICIT = 4
        UNKNOWN = 5

def _generate_md5(self, images: list[str]):
    log("> [nokufind.Post]: Generating md5 hashes.")
    if (type(images) != list or len(images) == 0):
        return []
    
    hashes = []

    for image in images:
        image_request = make_request(image)
        if (image_request.status_code > 400):
            hashes.append("")
            continue

        hashes.append(str(hashlib.md5(image_request.content).hexdigest()))

    self._Post__post_data["md5"] = hashes

class Post():
    """A wrapper around post data from multiple sources."""

    # Class variables.
    __rating_general = ["s", "safe", "general", "g"]
    __rating_sensitive = ["sensitive", "s"]
    __rating_questionable = ["questionable", "q"]
    __rating_explicit = ["e", "explicit"]

    # Max number of threads to be used when fetching the data.
    __max_threads = 10

    @staticmethod
    def set_max_threads(self, num_of_threads: int):
        """Sets the maximum number of threads to use when fetching data.

        Args:
            num_of_threads (int): The number of max threads.
        """
        Post.__max_threads = num_of_threads if num_of_threads > 0 else Post.__max_threads

    @staticmethod
    def _get_rating(rating_string: str) -> Rating:
        """[INTERNAL] Returns the corresponding Rating enum value.

        Args
        ~~~~
            rating_string (``str``): The rating string from the post data.

        Returns
        ~~~~~~~
            ``Rating``: The corresponding Rating enum value.
        """

        if (rating_string == None):
            return Rating.UNKNOWN

        rating_string = rating_string.lower()

        if (rating_string in Post.__rating_general):
            return Rating.GENERAL
        
        if (rating_string in Post.__rating_sensitive):
            return Rating.SENSITIVE
        
        if rating_string in Post.__rating_questionable:
            return Rating.QUESTIONABLE
        
        if rating_string in Post.__rating_explicit:
            return Rating.EXPLICIT
        
        return Rating.UNKNOWN
    
    def __init__(self, *, post_id: int, tags: list[str], sources: list[str], images: list[str], authors: list[str], source: str, preview: str, md5: list[str] | None, rating: str | None, parent_id: int | None, dimensions: list[tuple[int, int]], poster: str, poster_id: int, name: str):
        """
        Creates a new Post object. 

        If you are using any of the built-in Finders, you can use the to_post() static methods to automatically generate a Post object
        from the data object itself.

        Args:
            post_id (``int``): The ID of the post.
            tags (``list[str]``): A list containing all of the tags of the post.
            sources (``list[str]``): A list containing the sources of the images.
            images (``list[str]``): A list containing the URLs to the content (content can be anything, really).
            authors (``list[str]``): A list containing the authors of the work.
            source (``str``): A string representing the source from where the post data came from. (Ex: "danbooru", "konachan", etc...)
            preview (``str``): The URL of the image to be used as a preview.
            md5 (``str``): The md5 hash of the image. If None or empty, a list of md5 hashes will be automatically generated from the images.
            rating (``str | None``): The content rating. (Ex: "explicit", "general", etc...)
            parent_id (``int | None``): The ID of the Post's parent. If the Post has no parent, this should be None.
            dimensions (``list[tuple[int, int]]``): A list containing tuples containing the width and height of the images.
            poster (``str``): The username of the poster of the image.
            poster_id (``int``): The ID of the poster of the image.
            name (``str``): The title of the post, if it has any. If set to None, an automatic title will be generated.
        """

        # All post data is stored in a dictionary for easy conversion to a JSON file.
        self.__post_data = {}
        self.__post_data["post_id"] = post_id
        self.__post_data["tags"] = tags
        self.__post_data["sources"] = sources
        self.__post_data["images"] = images
        self.__post_data["authors"] = authors
        self.__post_data["source"] = source
        self.__post_data["preview"] = preview
        self.__post_data["md5"] = md5
        self.__post_data["rating"] = Post._get_rating(rating) if type(rating) != Rating else rating
        self.__post_data["parent_id"] = parent_id
        self.__post_data["poster"] = poster
        self.__post_data["poster_id"] = poster_id
        self.__post_data["filenames"] = [url.split("/")[-1] for url in self.__post_data["images"]]
        self.__post_data["name"] = f"Post #{post_id}" if not name else name
        self.__post_data["dimensions"] = dimensions

        # Some extra information that derives from the aforementioned post data and that should not be included in the exported JSON.
        self.__image = images[0]
        self.__is_video = self.__post_data["images"][0].endswith(".mp4")
        self.__is_zip = self.__post_data["images"][0].endswith(".zip")
        self.__parent = None
        self.__children = []
        self.__headers = {}
        self.__cookies = {}
        self.__data = []
        self.__fetched_data = False

        # If there is no md5 data, we generate it ourselves.
        if (type(self.__post_data["md5"]) != list or len(self.__post_data["md5"]) == 0 or self.__post_data["md5"][0] == None):
            threading.Thread(target=lambda: _generate_md5(self, self.__post_data["images"])).start()

    def __repr__(self):
        rep = Repr()
        param_string = ", ".join([f"{name}={rep.repr(value)}" for name, value in self.__post_data.items()])
        return f"<Post({param_string})>"
    
    def __str__(self):
        return json.dumps(self.__post_data)
    
    def __iter__(self):
        return self.__post_data.copy().__iter__()
    
    def __getitem__(self, key):
        return self.__post_data[key]
    
    def get_image_data(self, *, index: int = 0) -> bytes | None:
        """Requests and returns the data for the image at the given index.

        Args:
            index (``int``, optional): The number of the image to download, starting at 0. Defaults to 0.

        Returns:
            ``bytes | None``: Returns the data of the image, or None if the request failed.
        """
        image_url = self.images[index]
        image_request = make_request(image_url)

        if (image_request.status_code >= 400):
            log(f"> [nokufind.Post]: [{image_request.status_code}]: {image_request.text}")
            return None
        
        return image_request.content

    def download_item(self, path: str, *, index: int = 0) -> str:
        """Downloads and stores the image at the given index.

        Args:
            path (``str``): Directory where to store the downloaded image.
            index (``int``, optional): The number of the image, starting at 0. Defaults to 0.

        Returns:
            ``str``: The path to the downloaded file.
        """
        image_url = self.images[index]
        image_request = make_request(image_url, stream = True, headers = self.__headers, cookies = self.__cookies)

        if (image_request.status_code >= 400):
            log(f"[{image_request.status_code}]: {image_request.text}")
            return ""
        
        os.makedirs(path, exist_ok = True)

        content_size = int(image_request.headers.get("content-length", 1))
        progress = tqdm(total = content_size, unit = "iB", unit_scale = True)

        file_name = os.path.join(path, image_url.split("/")[-1])
        
        try:
            with open(file_name, "wb") as file:
                for data in image_request.iter_content(2048):
                    progress.update(len(data))
                    file.write(data)

            progress.close()
            return file_name
        except Exception as e:
            log(e)
            return ""

    def download_all(self, path: str) -> list[str]:
        """Downloads all of the images of the post.

        Args:
            path (``str``): Path to directory where to store all the files.

        Returns:
            ``list[str]``: List containing the paths to all of the downloaded files.
        """
        return [self.download_item(path, index = i) for i in range(len(self.images))]
    
    def fetch_data(self, only_image: bool = False) -> None:
        """Fetches the data of all of the post's images.

        Args:
            only_image (``bool``, optional): Whether to only download the main image. Defaults to False.
        """
        if (self.fetched_data and len(self.data) == len(self.images)):
            return
        
        session = requests.Session()

        if only_image:
            self.__data = [None]
            self._request_data(self.image, session, 0)
            self.fetched_data = True
            return

        self.__data = [None] * len(self.images)

        url_chunks = [self.images[i:i + Post.__max_threads] for i in range(0, len(self.images), Post.__max_threads)]

        with ThreadPoolExecutor(Post.__max_threads) as executor:
            last_index = 0
            for chunk in url_chunks:
                executor.map(self._request_data, chunk, [session] * len(chunk), list(range(last_index, len(chunk))))
                last_index = len(chunk)

    
    def _request_data(self, image_url: str, session: requests.Session, index: int) -> bool:
        """(Internal) Function used for fetching the data.
        ATTENTION! Do not use this function. Use ``fetch_data`` instead.

        Args:
            image_url (``str``): The URL of the image to fetch.
            session (``requests.Session``): A Session object used to make the request.
            index (``int``): The index where to store the downloaded data.

        Returns:
            ``bool``: Whether the image data was successfully fetched and stored or not.
        """
        response = session.get(image_url, headers = self.__headers, cookies = self.__cookies)

        if (response.status_code >= 400):
            log(f"> [nokufind.Post]: [{response.status_code}]: {response.text}")
            return False
        
        self.data[index] = response.content
        return True

    def _set_parent(self, parent_post: Post) -> Post:
        """(Internal) Used by ``post_get_parent`` to store the parent post.

        Args:
            parent_post (Post): The parent of the current post.

        Returns:
            ``Post``: Returns the parent post object.
        """
        self.__parent = parent_post
        return parent_post

    def _set_children(self, children_posts: list[Post]) -> list[Post]:
        """(Internal) Used by ``post_get_children`` to store the post's children.

        Args:
            children_posts (``list[Post]``): A list containing this post's children.

        Returns:
            ``list[Post]``: Returns the list of children posts.
        """
        self.__children = children_posts.copy()
        return children_posts
    
    def _set_cookies(self, cookies: dict) -> Post:
        """(Internal) Used to set the request cookies for downloading / fetching-related methods.

        Args:
            cookies (``dict``): The cookies to be used for the post's subsequent requests.

        Returns:
            ``Post``: Returns itself.
        """
        if (type(cookies) != dict):
            return self
        
        self.__cookies = cookies
        return self

    def _set_headers(self, headers: dict) -> Post:
        """(Internal) Used to set the request headers for downloading / fetching-related methods.

        Args:
            headers (``dict``): The headers to be used for the post's subsequent requests.

        Returns:
            ``Post``: Returns itself.
        """
        if (type(headers) != dict):
            return self
        
        self.__headers = headers
        return self

    @property
    def post_dict(self) -> dict:
        """Returns a copy of the post data dictionary for external use.

        Returns:
            dict: A dictionary containing all of the stored post data.
        """
        return self.__post_data.copy()

    @property
    def post_id(self) -> int:
        """The post ID.

        Returns:
            ``int``: The post's ID.
        """
        return self.__post_data["post_id"]
    
    @property
    def tags(self) -> list[str]:
        """The list of tags used for the image.

        Returns:
            ``list[str]``: A list containing the tags as strings.
        """
        return self.__post_data["tags"]

    @property
    def tag_string(self) -> str:
        """A string containing all of the tags.

        Returns:
            ``str``: Returns all of the tags in a single string (spaced with spaces).
        """
        return " ".join(self.__post_data["tags"])

    @property
    def sources(self) -> list[str]:
        """The sources of the images.
        In most cases, this is not the booru post URL, but the actual source from where the images came from.

        The data may or may not be actual valid URLs, since most boorus don't actually check that the images have valid sources.

        Returns:
            ``list[str]``: List containing all of the sources.
        """
        return self.__post_data["sources"]
    
    @property
    def images(self) -> list[str]:
        """Returns all of the images from the post.

        Returns:
            ``list[str]``: List containing the URLs of all of the images of the post.
        """
        return self.__post_data["images"]
    
    @property
    def image(self) -> str:
        """The first image of the post.

        Returns:
            str: The URL of the first image of the post.
        """
        return self.__image

    @property
    def authors(self) -> list[str]:
        """Returns the authors of the work.
        Usually the artists related to the work.

        Returns:
            ``list[str]``: List containing the authors of the work.
        """
        return self.__post_data["authors"]
    
    @property
    def source(self) -> str:
        """A string that identifies where the image came from. May be useful for using with the finder or for some external purposes.

        Returns:
            ``str``: A string that identifies where the image came from. (Ex: "danbooru", "pixiv", etc...)
        """
        return self.__post_data["source"]
    
    @property
    def preview(self) -> str:
        """Returns the post's preview image. 
        It is usually a downscaled version of the main image. Helpful for list previous of content.

        Returns:
            ``str``: URL to the preview image.
        """
        return self.__post_data["preview"]
    
    @property
    def md5(self) -> list[str]:
        """The md5 hashes of the images.
        Can be used for sorting and removing duplicate images, as well as used for finding the images in some boorus.

        Returns:
            list[str]: List containing the md5 hashes of the images.
        """
        return self.__post_data["md5"]
    
    @property
    def rating(self) -> Rating:
        """The rating of the image. (Ex: "general", "questionable", "explicit", etc...)

        Returns:
            ``Rating``: A Rating enum value representing the image's rating. 
        """
        return self.__post_data["rating"]
    
    @property
    def parent_id(self) -> int | None:
        """Returns the post's parent ID, if it has one.

        Returns:
            int | None: The ID of the post's parent, or None if the post does not have a parent.
        """
        return self.__post_data["parent_id"]
    
    @property
    def parent(self) -> Post | None:
        """Returns the post's parent, if it has one and has been searched for using ``post_get_parent``.

        Returns:
            ``Post | None``: A Post object containing the post's parent data.
        """
        return self.__parent
    
    @property
    def children(self) -> list[Post]:
        """Returns the post's children, if it has any and have been searched for using ``post_get_children``.

        Returns:
            ``list[Post]``: A list containing the post's children as Post objects.
        """
        return self.__children
    
    @property
    def filenames(self) -> list[str]:
        """The filenames of every image as stored online. Usually the md5 hash but can vary.

        Returns:
            ``list[str]``: List containing the filenames of every image.
        """
        return self.__post_data["filenames"]
    
    @property
    def name(self) -> str:
        """The name of the post, for the sources that support it.

        Returns:
            ``str``: The name of the post, or an auto-generated one.
        """
        return self.__post_data["name"]
    
    @property
    def dimensions(self) -> list[tuple[int, int]]:
        """The dimensions of all of the post's images.

        Returns:
            ``list[tuple[int, int]]``: A list containing tuples with the images' width and height as integers.
        """        
        return self.__post_data["dimensions"]
    
    @property
    def poster(self) -> str:
        """The name of the user who posted the images.

        Returns:
            ``str``: A string containing the poster's username.
        """
        return self.__post_data["poster"]
    
    @property
    def poster_id(self) -> int:
        """The ID of the user who posted the images.

        Returns:
            ``int``: The ID for the user who posted the images, as stored in the original source.
        """
        return self.__post_data["poster_id"]
    
    @property
    def is_video(self) -> bool:
        """Returns whether the post content is an MP4 video or not.

        Returns:
            ``bool``: ``True`` if the post content is an MP4 video.
        """
        return self.__is_video
    
    @property
    def is_zip(self) -> bool:
        """Returns whether the post content is a ZIP file.

        Returns:
            ``bool``: ```True`` if the post content is a ZIP file.
        """
        return self.__is_zip
    
    @property
    def data(self) -> list[bytes]:
        """Returns a list containing the raw fetched data of all the images.

        This list is empty by default. To fill it, use the ``fetch_data`` method.

        Returns:
            ``list[bytes]``: List containing raw content data as bytes.
        """
        return self.__data
    
    @property
    def fetched_data(self) -> bool:
        """Whether the data has been fetched or not.

        Returns:
            ``bool``: ``True`` if the data has been successfully fetched.
        """
        return self.__fetched_data


"""
    Finder.py

    Nokutoka Momiji

    Contains the Finder class that acts as a wrapper around multiple booru / image searching APIs. 
    It allows you to get posts, comments, notes, etc... from various sources. 
    
    The Finder class provides methods for adding and removing subfinders, searching for posts, comments, and notes, 
    downloading images, and more.

    This script utilizes asyncio and threading to handle asynchronous operations and parallel downloading of images. 
    It also makes use of third-party libraries such as httpx and aiometer for HTTP requests and asynchronous processing.
"""

# System Modules
import os
import io
import asyncio
import functools
import threading
from random import shuffle

import httpx
import aiometer

from nokufind import Post, Comment, Note
from nokufind.Subfinder import (ISubfinder,
                                SubfinderConfiguration, 
                                DanbooruFinder, 
                                Rule34Finder,
                                KonachanFinder,
                                YandereFinder,
                                GelbooruFinder,
                                PixivFinder)

from nokufind.Utils import log, split

_list_lock = threading.Lock()

class Finder():
    """
    A wrapper around multiple booru / image searching APIs that allows you to get posts, comments, notes, etc.
    """

    def __init__(self):
        self.__clients: dict[str, ISubfinder] = {}
        self.__config: SubfinderConfiguration = SubfinderConfiguration()
        self.__name = f"nokufind.Finder"

    def add_subfinder(self, name: str, subfinder: ISubfinder) -> None:
        """Adds a subfinder to the list of clients to use in subsequent functions.

        Args:
            name (``str``): The name that will be used to identify the subfinder.
            subfinder (``ISubfinder``): The subfinder to be stored and used.

        Raises:
            ``TypeError``: Raised if the provided subfinder doesn't inherit from ``ISubfinder``.
        """
        if (not isinstance(subfinder, ISubfinder)):
            raise TypeError("Subfinder must inherit from ISubfinder.")
        self.__clients[name] = subfinder

    def remove_subfinder(self, name: str) -> None:
        """Removes a subfinder.

        Args:
            name (str): The name of the subfinder to remove.
        """
        self.__clients.pop(name)

    def get_subfinder(self, name: str) -> ISubfinder | None:
        """Returns a subfinder with the given name, if it has one.

        Args:
            name (``str``): The name of the subfinder to look for.

        Returns:
            ``ISubfinder`` | None: The subfinder instance, or None if none was found.
        """
        return self.__clients.get(name, None)
    
    def has_subfinder(self, name: str) -> bool:
        """Checks whether the finder contains a subfinder with the given name.

        Args:
            name (``str``): The subfinder to look for.

        Returns:
            ``bool``: Returns ``True`` if it contains a subfinder with the given name.
        """
        return (name in self.__clients.keys())
    
    def add_default(self, *, danbooru_key = ""):
        """Adds a few built-in subfinders.

        Args:
            danbooru_key (``str``, optional): Key for the Danbooru API. Defaults to "".
        """
        self.add_subfinder("danbooru", DanbooruFinder(danbooru_key))
        self.add_subfinder("rule34", Rule34Finder())
        self.add_subfinder("konachan", KonachanFinder())
        self.add_subfinder("yande.re", YandereFinder())
        self.add_subfinder("gelbooru", GelbooruFinder())

    def search_posts(self, tags: str | list[str] = "", *, limit: int = 100, page: int | None = None, client: str | None = None) -> list[Post]:
        """Searches for posts with the given tags.

        Args:
            tags (``str | list[str]``, optional): The tags to match for. Defaults to "".
            limit (``int``, optional): The maximum number of posts to look for per client. Defaults to 100.
            page (``int | None``, optional): The page number. Defaults to None.
            client (``str | None``, optional): The name of the client to use. If None, all clients are used. Defaults to None.

        Returns:
            list[Post]: List containing all the found posts.
        """
        if type(client) == str:
            self._check_subfinder_exists(client)
            return self.__clients[client].search_posts(tags = tags, limit = limit, page = page)
        
        all_posts: list[Post] = []

        for name, subfinder in self.__clients.items():
            posts = subfinder.search_posts(tags = tags, limit = limit, page = page)
            if len(posts) == 0:
                log(f"> [{self.__name}]: Subfinder {name} returned 0 posts.")
            all_posts += posts

        return all_posts
    
    def get_post(self, post_id: int, *, client: str | None = None) -> Post | None:
        """Gets a post with the given ID.

        Args:
            post_id (``int``): ID of the post to look for.
            client (``str | None``, optional): The name of the client to use. If ``None``, all clients will be used. Defaults to None.

        Returns:
            ``Post | None``: Post object containing all the data, or None if no post was found.
        """
        if type(client) == str:
            self._check_subfinder_exists(client)
            return self.__clients[client].get_post(post_id)
        
        for subfinder in self.__clients.values():
            post = subfinder.get_post(post_id = post_id)

            if post != None:
                return post
            
        return None
    
    def search_comments(self, *, client: str | None = None, post_id: int | None = None, limit: int = 100, page: int | None = None) -> list[Comment]:
        """Searches for comments.

        Args:
            client (``str | None``, optional): The subfinder to use. If None, all subfinders will be used. Defaults to None.
            post_id (``int | None``, optional): The ID of the post to get the comments from. Defaults to None.
            limit (``int``, optional): The maximum number of comments to return per client. Defaults to 100.
            page (``int | None``, optional): The page number. Defaults to None.

        Returns:
            list[Comment]: List containing all the found comments.
        """
        if type(client) == str:
            self._check_subfinder_exists(client)
            return self.__clients[client].search_comments(post_id = post_id, limit = limit, page = page)
        
        all_comments = []

        for subfinder in self.__clients.values():
            comments = subfinder.search_comments(post_id = post_id, limit = limit, page = page)
            all_comments += comments

        return all_comments

    def get_comment(self, comment_id: int, *, client: str | None = None) -> Comment | None:
        """Returns a comment with the given ID.

        Args:
            comment_id (``int``): The ID of the comment to look for.
            client (``str | None``, optional): The subfinder to use. If None, all subfinders will be used. Defaults to None.

        Returns:
            ``Comment | None``: A Comment object containing the data, or None if no comment was found.
        """
        if type(client) == str:
            self._check_subfinder_exists(client)
            return self.__clients[client].get_comment(comment_id)
        
        for subfinder in self.__clients.values():
            comment = subfinder.get_comment(comment_id)
            
            if comment != None:
                return comment
            
        return None
    
    def get_notes(self, post_id: int, *, client: str | None = None) -> list[Note]:
        """Gets the notes of a post.

        Args:
            post_id (``int``): The ID of the post to get the notes from.
            client (``str | None``, optional): The subfinder to use. If None, all subfinders will be used. Defaults to None.

        Returns:
            ``list[Note]``: A list containing the notes.
        """
        if type(client) == str:
            self._check_subfinder_exists(client)
            return self.__clients[client].get_notes(post_id)
        
        all_notes = []
        for subfinder in self.__clients.values():
            notes = subfinder.get_notes(post_id)
            all_notes += notes

        return all_notes

    def post_get_parent(self, post: Post, *, client: str | None = None) -> Post:
        """Gets the parent post of a post.

        Args:
            post (``Post``): The post to find the parent of.
            client (``str | None``, optional): The subfinder to use. If None, all subfinders will be used. Defaults to None.

        Returns:
            ``Post``: The parent post of the post.
        """
        if type(client) == str:
            self._check_subfinder_exists(client)
            return self.__clients[client].post_get_parent(post)
        
        for subfinder in self.__clients.values():
            parent_post = subfinder.post_get_parent(post)

            if parent_post != None:
                return parent_post
            
        return None

    def post_get_children(self, post: Post, *, client: str | None = None) -> list[Post]:
        """Gets the children posts of the a post.

        Args:
            post (``Post``): The post to find the children of.
            client (``str | None``, optional): The subfinder to use. If None, all subfinders are used. Defaults to None.

        Returns:
            list[Post]: _description_
        """
        if type(client) == str:
            self._check_subfinder_exists(client)
            return self.__clients[client].post_get_children(post)
        
        all_children = []

        for subfinder in self.__clients.values():
            children_post = subfinder.post_get_children(post)
            all_children += children_post

        return all_children
            
    def download_fast(self, post_list: list[Post], directory: str) -> list[str]:
        """Downloads all of the images in a list of posts to the provided directory.

        Args:
            post_list (``list[Post]``): A list containing all of the posts to download.
            directory (``str``): A path to a directory. If the path doesn't exist, it will be created.

        Raises:
            ``TypeError``: Raised if post_list is not a list.

        Returns:
            ``list[str]``: List containing the paths to all of the downloaded files.
        """
        return asyncio.run(self.download_fast_async(post_list, directory))

    async def download_fast_async(self, post_list: list[Post], directory: str) -> list[str]:
        """Downloads all of the images in a list of posts to the provided directory.

        Args:
            post_list (``list[Post]``): A list containing all of the posts to download.
            directory (``str``): A path to a directory. If the path doesn't exist, it will be created.

        Raises:
            ``TypeError``: Raised if post_list is not a list.

        Returns:
            ``list[str]``: List containing the paths to all of the downloaded files.
        """
        if (type(post_list) != list):
            raise TypeError("Value should be a list of Post objects. Try using search_posts().")
        
        os.makedirs(directory, exist_ok = True)

        # In an attempt to minimize the amount of constant requests to a single source
        # we create a shuffled version of the list. If there are posts of multiple sources,
        # this could theoretically lower the amount of requests to a single source at a time.
        temp_post_list = post_list.copy()
        shuffle(temp_post_list)

        client = httpx.AsyncClient()
        request_function = functools.partial(self._download_and_safe, client, directory)
        end_paths = []

        async with aiometer.amap(request_function, post_list, max_at_once = 10, max_per_second = 10) as results:
            async for result in results:
                end_paths += result

        return end_paths

    async def _download_and_safe(self, client: httpx.AsyncClient, directory: str, post: Post) -> list[str]:
        """(Internal) Function for doing the actual downloading. It creates threads for each image in a post.
        ATTENTION! You should not use this function, use ``download_fast`` or ``download_fast_async`` instead.

        Args:
            client (``httpx.AsyncClient``): An Asynchronous Client to use for making the requests.
            directory (``str``): The directory where the images will be stored in.
            post (``Post``): The Post object with the images to download.

        Returns:
            ``list[str]``: List containing the paths of the downloaded images.
        """
        def store_response(list_ref: list, directory: str, filename: str, response: httpx.Response):
            """(Internal) Handles the request for an image, writes the content to a file, and stores the path to the file
            in the given list, using a lock to make sure that it is the only one writing to it.

            Args:
                list_ref (list): The list where to store the path to the downloaded file.
                directory (str): The directory where to store the file.
                filename (str): The filename of the file.
                response (httpx.Response): The response object from the request.
            """
            if response.status_code >= 400:
                log(f"> [{self.__name}]: Request returned {response.status_code} for \"{response.url}\".")
                return
            
            file_path = os.path.join(directory, filename)
            with open(file_path, "wb") as f:
                f.write(response.content)

            with _list_lock:
                list_ref.append(file_path)

        end_paths: list[str] = []
        threads: list[threading.Thread] = []
        
        # For each image request, we create a thread to handle storing the file.
        for index, image in enumerate(post.images):
            response = await client.get(image, timeout = None, headers = self.configuration.headers)
            thread = threading.Thread(target = lambda: store_response(end_paths, directory, post.filenames[index], response))
            threads.append(thread)
            thread.start()

        for thread in threads:
            thread.join()

        return end_paths
            
    def on_config_change(self, key: str, value, is_cookie: bool, is_header: bool):
        """(Internal) Callback function for when the configuration settings are modified.

        Args:
            key (str): The key that was changed.
            value (_type_): The value to which the key was changed to.
            is_cookie (bool): Whether the key was a request cookie.
            is_header (bool): Whether the key was a request header.
        """
        for client in self.__clients.values():
            client_method = client.configuration.set_cookie if is_cookie else client.configuration.set_header
            client_method = client.configuration.set_header if is_header else client_method

            client_method(key, value)

    def _check_subfinder_exists(self, client: str):
        """(Internal) Simply checks if the subfinder is stored in the clients list.

        Args:
            client (str): The name of the subfinder.

        Raises:
            ``ValueError``: Raised if there is no Subfinder with the given client name.
        """
        if not client in self.__clients.keys():
            raise ValueError(f"Finder contains no Subfinder named \"{client}\".")
        
    @property
    def finders(self) -> list[ISubfinder]:
        """Returns the list of subfinders.

        Returns:
            ``list[ISubfinder]``: A list containing the subfinders stored within this finder instance.
        """
        return self.__clients
    
    @property
    def configuration(self) -> SubfinderConfiguration:
        """Returns the configuration object for the finder, which can be used to set configuration for all subfinders, as well
        as request headers and cookies.

        All modifications to this configuration object will apply to all subfinders. If you wish to modify the configuration for a
        specific subfinder, you can use the ``get_subfinder`` method to access the subfinder's configuration object directly.

        Returns:
            ``SubfinderConfiguration:`` A configuration object used by the finder. 
        """
        return self.__config


import requests
import xml.etree.ElementTree as ET
from math import ceil

_cookies = {
    "cf_clearance": "4PtNZd5PGSDKDBXhASc1z_mC.tKc.ggMrvxhx2s3GdE-1709327262-1.0.1.1-azAPC335pQo4d9v.0TqPligg5htX.RtlAn36lYJx1Vc9IHvYDWrWjSBCaxciXtfGeawADZZ9MDQG2c7iXT27vA"
}

_headers = {
    "Referer": "https://app-api.pixiv.net/"
}

USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/121.0.0.0 Safari/537.36 OPR/107.0.0.0"
PIXIV_REFERER = "https://app-api.pixiv.net/"

should_log = False

def make_request(url: str, params = None, *, post: bool = False, cookies = _cookies, headers = _headers, stream: bool = False):
    request_function = requests.get if not post else requests.post
    return request_function(url, params = params, headers = headers, cookies = cookies, stream = stream)

def parse_xml(xml_string: str):
    # Parse the XML string
    root = ET.fromstring(xml_string)

    # Convert XML to a Python dictionary
    def parse_element(element):
        result = {}
        for child in element:
            if child:
                child_data = parse_element(child)
                if child.tag in result:
                    if type(result[child.tag]) is list:
                        result[child.tag].append(child_data)
                    else:
                        result[child.tag] = [result[child.tag], child_data]
                else:
                    result[child.tag] = child_data
            else:
                result[child.tag] = child.text
        return result

    return parse_element(root)

def attempt_conversion(item, new_type):
    try:
        return new_type(item)
    except:
        return item
    
def assert_conversion(item, new_type, item_name: str = ""):
    item = attempt_conversion(item, new_type)
    item_name = "Element" if len(item_name) == 0 else item_name

    if (type(item) != new_type):
        raise TypeError(f"{item_name} must be {new_type}, not {type(item)}.")

    return item
    
def log(content):
    if not should_log:
        return
    print(content)

def get(list_ref: list, index: int, default_value = None):
    if (index >= len(list_ref)):
        return default_value
    
    return list_ref[index]

def parse_tags(tags):
    current_text = ""
    tag_list = []
    num_of_parenthesis = 0
    
    tags = tags if type(tags) == str else " ".join(tags)

    for index, char in enumerate(tags):
        if char == " " and num_of_parenthesis == 0:
            tag_list.append(current_text)
            current_text = ""
            continue

        num_of_parenthesis += 1 if char == '(' else (-1 if char == ')' else 0)

        if num_of_parenthesis < 0:
            raise SyntaxError(f"Unmatched closing parenthesis found in tags.\n    {tags}\n    {'~' * index}^")

        current_text += char
    
    tag_list.append(current_text)

    return tag_list

def expand(list_ref, new_size, value = None):
    current_len = len(list_ref)

    if (new_size <= current_len):
        return
    
    difference = new_size - current_len

    for _ in range(difference):
        list_ref.append(value)

def split(list_ref, num_of_splits):
    split_lists = []
    max_num_of_items = ceil(len(list_ref) / num_of_splits)

    current_list = []
    current_list_count = 0

    for item in list_ref:
        current_list.append(item)
        current_list_count += 1

        if current_list_count >= max_num_of_items:
            split_lists.append(current_list)
            current_list = []
            current_list_count = 0

    return split_lists




from typing import cast

from enma import Enma, DefaultAvailableSources, CloudFlareConfig, NHentai, Manga
from enma.domain.entities.manga import Chapter

from nokufind import Post, Comment, Note
from nokufind.Subfinder import ISubfinder, SubfinderConfiguration
from nokufind.Utils import assert_conversion, get, log, USER_AGENT
from nokufind.Post import Rating

class NHentaiFinder(ISubfinder):
    @staticmethod
    def to_post(post_data: Manga) -> Post:
        first_page = post_data.chapters[0].pages[0]
        return Post(
            post_id = int(post_data.id),
            tags = [genre.name for genre in post_data.genres],
            sources = [f"https://nhentai.net/g/{post_data.id}/"],
            images = [page.uri for page in post_data.chapters[0].pages],
            authors = [author.name for author in post_data.authors],
            source = "nhentai",
            preview = post_data.thumbnail.uri,
            md5 = None,
            rating = Rating.EXPLICIT,
            parent_id = None,
            dimensions = [(page.width, page.height) for page in post_data.chapters[0].pages],
            poster = "",
            poster_id = None,
            name = post_data.title.english
        )
    
    @staticmethod
    def to_comment(comment_data) -> Comment:
        return None
    
    @staticmethod
    def to_note(note_data) -> Note:
        return None

    def __init__(self, cf_clearance: str = "") -> None:
        self.__client = Enma[DefaultAvailableSources]()
        self.__client.source_manager.set_source("nhentai")

        self.__cf_config = CloudFlareConfig(USER_AGENT, cf_clearance)
        self.__source = cast(NHentai, self.__client.source_manager.source)
        self.__source.set_config(self.__cf_config)

        self.__config = SubfinderConfiguration(self.on_client_change)
        self.__config.set_header("Referer", "https://nhentai.net")

        self.__name = f"nokufind.Subfinder.{self.__class__.__name__}"

        if (len(cf_clearance) == 0):
            log(f"> [{self.__name}]: This Finder requires a valid cf_clearance cookie.")
            log(f"> [{self.__name}]: Please set the cookie via \"configuration.set_cookie(\"cf_clearance\", [cookie value here])\"")

    def search_posts(self, tags: str | list[str], *, limit: int = 100, page: int | None = None) -> list[Post]:
        self._check_client()

        if type(tags) == list:
            tags = " ".join(tags)

        limit = abs(assert_conversion(limit, int, "limit"))
        current_page = 1 if page == None else page
        posts = []
        EXPECTED_SIZE = 25
        previous_size = EXPECTED_SIZE
        oversized = False

        while (previous_size == EXPECTED_SIZE):
            raw_results = self.__client.search(tags, current_page).results
            previous_size = len(raw_results)

            posts += [self.get_post(post.id) for post in raw_results]

            if (len(posts) >= limit):
                oversized = True
                break

            current_page += 1

        if (oversized):
            return posts[:limit]
        
        return posts

    def get_post(self, post_id: int) -> Post | None:
        self._check_client()

        post_id = str(post_id)

        manga = self.__client.get(post_id)

        return NHentaiFinder.to_post(manga)._set_headers(self.configuration.headers) if manga != None else None

    def search_comments(self, *, post_id=None, limit=None, page=None) -> list[Comment]:
        return super().search_comments(post_id=post_id, limit=limit, page=page)

    def get_comment(self, comment_id: int) -> Comment | None:
        return super().get_comment(comment_id)
    
    def get_notes(self, post_id: int) -> list[Note]:
        return []

    def post_get_parent(self, post: Post) -> Post | None:
        return None
    
    def post_get_children(self, post: Post) -> list[Post]:
        return []

    def on_client_change(self, key: str, value, is_cookie, is_header):
        if (is_cookie and key == "cf_clearance"):
            self.__cf_config.cf_clearance = value
            self.__source.set_config(self.__cf_config)

    def _check_client(self):
        if (not isinstance(self.__client, Enma)):
            raise RuntimeError(f"Client must be an Enma instance, not {type(self.__client)}.")
        
        if (len(self.__cf_config.cf_clearance) == ""):
            log(f"> [{self.__name}]: This Finder requires a valid cf_clearance cookie.")
            log(f"> [{self.__name}]: Please set the cookie via \"configuration.set_cookie(\"cf_clearance\", [cookie value here])\"")

    @property
    def configuration(self):
        return self.__config
        



from datetime import datetime
from itertools import combinations
from time import sleep

from pybooru import Danbooru
from pybooru.exceptions import PybooruHTTPError

from nokufind import Post, Comment, Note
from nokufind.Subfinder import ISubfinder, SubfinderConfiguration
from nokufind.Utils import attempt_conversion, log, get, parse_tags

def _post_is_valid(post: dict) -> bool:
    return (post.get("md5") != None and post.get("file_url") != None)

def _comment_is_valid(comment: dict) -> bool:
    return not comment.get("is_deleted")

def _filter_invalid_posts(posts: list[dict]) -> list[dict]:
    return [post for post in posts if _post_is_valid(post)]

def _filter_invalid_comments(comments: list[dict]) -> list[dict]:
    return [comment for comment in comments if _comment_is_valid(comment)]

class DanbooruFinder(ISubfinder):
    # Static methods.
    @staticmethod
    def to_post(post_data: dict) -> Post:
        """Creates a Post object from Danbooru data.

        Args
        ~~~~
            post_data (``dict``): A dictionary containing all of the Danbooru post data.

        Returns
        ~~~~~~~
            ``Post``: A Post object containing the extracted data.
        """

        if "author" not in post_data.keys():
            post_data["author"] = f"User {post_data['uploader_id']}"

        return Post(
            post_id = post_data["id"],
            tags = post_data["tag_string"].split(" "),
            sources = post_data["source"].split(" "),
            images = [post_data["file_url"]],
            authors = post_data["tag_string_artist"].split(" "),
            source = "danbooru",
            preview = post_data["preview_file_url"],
            md5 = [post_data["md5"]],
            rating = post_data["rating"],
            parent_id = post_data["parent_id"],
            dimensions = [(post_data["image_width"], post_data["image_height"])],
            poster = post_data["author"],
            poster_id = post_data["uploader_id"],
            name = None
        )
    
    @staticmethod
    def to_comment(comment_data) -> Comment:
        return Comment(
            comment_id = comment_data["id"],
            post_id = comment_data["post_id"],
            creator_id = comment_data["creator_id"],
            creator = "",
            body = comment_data["body"],
            source = "danbooru",
            created_at = datetime.strptime(comment_data["created_at"], "%Y-%m-%dT%H:%M:%S.%f%z")
        )

    @staticmethod
    def to_note(note_data: dict) -> Note:
        """Creates a Note object from Danbooru note data.

        Args:
            note_data (``dict``): A dictionary containing the Danbooru note data.

        Returns:
            ``Note``: A Note object containing the extracted data.
        """

        return Note(
            note_id = note_data["id"],
            created_at = datetime.strptime(note_data["created_at"], "%Y-%m-%dT%H:%M:%S.%f%z"),
            x = note_data["x"],
            y = note_data["y"],
            width = note_data["width"],
            height = note_data["height"],
            body = note_data["body"],
            source = "danbooru",
            post_id = note_data["post_id"]
        )
    
    def __init__(self, api_key = ""):
        self.__client = Danbooru("danbooru", api_key = api_key)

        self.__config = SubfinderConfiguration(self.on_config_change)
        self.__config._set_property("api_key", api_key)

        self.__page = 1
        self.__comments_page = 1

        self.__name = f"nokufind.Subfinder.{self.__class__.__name__}"

    def _get_all_posts(self, tags: str, limit: int = 100, page: int | None = None) -> list[Post]:
        current_posts = []
        current_size = 100
        current_page = page if type(page) == int else 1
        
        while (current_size == 100):
            try:
                raw_posts = self.__client.post_list(tags = tags, limit = 100, page = current_page)
            except KeyError:
                break
            except PybooruHTTPError as e:
                log(f"> [{self.__name}]: PybooruHTTPError: {e}")
                if "410" in e._msg:
                    break
                else:
                    sleep(0.1)
                    continue

            current_size = len(raw_posts)

            raw_posts = _filter_invalid_posts(raw_posts)
            posts = [DanbooruFinder.to_post(post) for post in raw_posts]

            current_posts += posts

            if len(current_posts) >= limit:
                break

            current_page += 1

        current_posts.sort(key=lambda x: x.post_id)

        if len(current_posts) > limit:
            return current_posts[:limit]
        
        return current_posts

    def _get_multiple_tags(self, tags: str | list[str], limit: int = 100, page: int | None = None) -> list[Post]:
        parsed_tags = parse_tags(tags)
        
        num_of_tags = len(parsed_tags)
        
        if num_of_tags <= 2:
            tag_string = " ".join(parsed_tags)
            return self._get_all_posts(tag_string, limit = limit, page = page)
        
        total_posts: list[Post] = []
        filtered_posts: list[Post] = []

        for index, tag in enumerate(parsed_tags):
            if tag == "or":
                s1 = []
                s2 = []
                prev_tags: str = get(parsed_tags, index - 1, None)
                next_tags: str = get(parsed_tags, index + 1, None)

                if prev_tags:
                    s1 = self._get_multiple_tags(prev_tags.strip("()"), limit = limit, page = page)
                if next_tags:
                    s2 = self._get_multiple_tags(next_tags.strip("()"), limit = limit, page = page)
                return s1 + s2
            elif "or" in tag:
                return self._get_multiple_tags(next_tags.strip("()"), limit = limit, page = page)
        
        if (num_of_tags % 2 == 0):
            total_combinations = [" ".join(total_posts[i:i + 2]) for i in range(num_of_tags, step = 2)]
        else:
            total_combinations = list(combinations(parsed_tags, 2))

        for combination in total_combinations:
            tag_string = " ".join(combination)
            log(f"> [{tag_string}]: \"{tag_string}\".")

            current_posts = self._get_all_posts(tag_string, limit, page)

            if (len(current_posts)):
                log(f"> [{self.__name}]: No posts found.")
                return []
            
            total_posts += current_posts

        for post in total_posts:
            add_to_list = True
            for tag in parsed_tags:
                if tag not in post.tags and not tag.startswith("rating:"):
                    add_to_list = False
                    break
            
            if add_to_list:
                filtered_posts.append(post)

        if (len(filtered_posts) > limit):
            return filtered_posts[:limit]
        
        return filtered_posts

    def search_posts(self, tags: str | list[str], *, limit: int = 100, page: int | None = None) -> list[Post]:
        self._check_client()

        # Danbooru takes in tags as a string, so we need to join them together.
        if (type(tags) == list):
            tags = " ".join()

        if page == None:
            page = self.__page
        else:
            self.__page = page

        return self._get_multiple_tags(tags = tags, limit = limit, page = page)
    
    def get_post(self, post_id: int) -> Post | None:
        self._check_client()

        # Post ID must be an int, so we convert it or raise an error if the conversion fails.
        post_id = attempt_conversion(post_id, int)

        if (type(post_id) != int):
            raise TypeError(f"post_id should be an int, not {type(post_id)}")

        # Danbooru flat out raises a PybooruHTTPError when it doesn't find a post, so we deal with that.
        try:
            raw_post = self.__client.post_show(post_id)
            
            # Normally, I would do this using a ternary operator, but the to_post function fails if data is not valid.
            if (not _post_is_valid(raw_post)):
                return None
            
            return DanbooruFinder.to_post(raw_post)
        except PybooruHTTPError as e:
            log(f"> [{self.__name}]: Danbooru raised an HTTP error.\n{e}")
            return None
        
    def search_comments(self, *, post_id = None, limit = None, page = None):
        raw_comments = self.__client.comment_list("comment", post_id = post_id, limit = limit, page = page)
        raw_comments = _filter_invalid_comments(raw_comments)

        return [DanbooruFinder.to_comment(comment) for comment in raw_comments if comment != None]
    
    def get_comment(self, comment_id: int) -> Comment | None:
        self._check_client()

        comment_id = attempt_conversion(comment_id, int)

        if (type(comment_id) != int):
            raise TypeError(f"comment_id should be an int, not {type(comment_id)}.")

        try:
            raw_comment = self.__client.comment_show(comment_id)

            return DanbooruFinder.to_comment(raw_comment)
        except PybooruHTTPError as e:
            log(f"> [{self.__name}]: Danbooru raised an HTTP error.\n{e}")
            return None

    def get_notes(self, post: int | Post) -> list[Note]:
        self._check_client()

        post_id = post.post_id if type(post) == Post else post
        post_id = attempt_conversion(post_id, int)

        if (type(post) != int):
            raise TypeError(f"post_id should be an int or a Post object, not {type(post_id)}")
        
        raw_notes = self.__client.note_list(post_id = post_id)

        return [DanbooruFinder.to_note(note) for note in raw_notes if note != None]
    
    def post_get_parent(self, post: Post) -> Post | None:
        if (not isinstance(post, Post)):
            raise TypeError(f"post should be a Post object, not {type(post)}")
        
        if (not post.parent_id):
            return None

        return post._set_parent(self.get_post(post.parent_id))

    def post_get_children(self, post: Post) -> list[Post]:
        if (not isinstance(post, Post)):
            raise TypeError(f"post should be a Post object, not {type(post)}")
        
        children_posts = self.search_posts(f"parent:{post.post_id}")

        for index, child_post in enumerate(children_posts):
            if child_post.post_id == post.post_id:
                children_posts.pop(index)
                break

        return post._set_children(children_posts)

    def next_page(self):
        self.__page += 1

    def on_config_change(self, key: str, value, is_cookie: bool, is_header: bool):
        if (key == "api_key"):
            self.__client = Danbooru("danbooru", api_key = value)

    def _check_client(self):
        if (not isinstance(self.__client, Danbooru)):
            raise RuntimeError(f"Client must be of type Danbooru, not {type(self.__client)}.")
        
    @property
    def configuration(self):
        return self.__config


from datetime import datetime

from rule34Py import Post as r34Post
from rule34Py.post_comment import PostComment as r34Comment

from nokufind import Post, Comment, Note
from nokufind.Subfinder import ISubfinder, SubfinderConfiguration
from nokufind.Utils import Rule34API, assert_conversion, parse_tags

class Rule34Finder(ISubfinder):
    @staticmethod
    def to_post(post_data: r34Post) -> Post:
        """Creates a Post object from rule34 post data.

        Args
        ~~~~
            post_data (``rule34Py.post.Post``): A rule34 Post object (from rule34Py) containing all of the post data.

        Returns
        ~~~~~~~
            ``Post``: A Post object containing the extracted data.
        """

        # The Finder uses a custom Rule34API object which means posts contain some extra data
        # not found originally in rule34Py's Post object, so I check for that if the user attempts
        # to use a post from rule34Py.
        if (not hasattr(post_data, "source")):
            post_data.source = ""

        if (not hasattr(post_data, "content")):
            post_data.content = post_data.video if post_data.content_type == "video" else post_data.image

        if (not hasattr(post_data, "parent_id")):
            post_data.parent_id = None

        return Post(
            post_id = post_data.id,
            tags = post_data.tags,
            sources = [post_data.source.split(" ")], #post_data.source.split(" "),
            images = [post_data.content],
            authors = [""],
            source = "rule34",
            preview = post_data.sample,
            md5 = [post_data.hash],
            rating = post_data.rating,
            parent_id = post_data.parent_id,
            dimensions = [(post_data.size[0], post_data.size[1])],
            poster = post_data.owner,
            poster_id = -1,
            name = None
        )
    
    @staticmethod
    def to_comment(comment_data: r34Comment) -> Comment:
        """Creates a Comment object from rule34 comment data.

        Args:
            comment_data (``rule34Py.post_comment.PostComment``): A PostComment object (from rule34Py) containing all of the comment data.

        Returns:
            ``Comment``: A Comment object containing the extracted data.
        """
        
        if (not hasattr(comment_data, "creator")):
            comment_data.creator = "Unknown"
        
        return Comment(
            comment_id = comment_data.id,
            post_id = comment_data.post_id,
            creator_id = comment_data.author_id,
            creator = comment_data.creator,
            body = comment_data.body,
            source = "rule34",
            created_at = datetime.strptime(comment_data.creation, "%Y-%m-%d %H:%M")
        )
    
    @staticmethod
    def to_note(note_data: dict) -> Note:
        return Note(
            note_id = note_data["id"],
            created_at = datetime.strptime(note_data["created_at"], "%Y-%m-%d %H:%M:%S"),
            x = note_data["x"],
            y = note_data["y"],
            width = note_data["width"],
            height = note_data["height"],
            body = note_data["body"],
            source = "rule34",
            post_id = note_data["post_id"]
        )

    def __init__(self):
        self.__client = Rule34API()
        self.__config = SubfinderConfiguration()

    def _get_all_posts(self, tags: list[str], limit: int, page: int | None) -> list[r34Post]:
        all_posts = []
        current_page = page if page != None else 0
        previous_size = 1000
        oversized = False

        while (previous_size == 1000):
            raw_posts = self.__client.search(tags, current_page)
            previous_size = len(raw_posts)

            all_posts += raw_posts

            if len(all_posts) >= limit:
                oversized = True
                break

            current_page += 1

        if oversized:
            return all_posts[:limit]
        
        return all_posts
    
    def _get_all_comments(self, page: int | None, limit: int) -> list[r34Comment]:
        all_comments = []
        current_page = page if page != None else 1
        previous_size = 100
        oversized = False

        while (previous_size == 100):
            raw_comments = self.__client.search_comments(current_page)
            previous_size = len(raw_comments)

            all_comments += raw_comments

            if len(all_comments) >= limit:
                oversized = True
                break

            current_page += 1

        if oversized:
            return all_comments[:limit]
        
        return all_comments

    def search_posts(self, tags: str | list[str], *, limit: int = 100, page: int | None = None) -> list[Post]:
        self._check_client()
        
        tags = parse_tags(tags) if type(tags) != list else tags

        raw_posts = self._get_all_posts(tags, limit, page)

        return [Rule34Finder.to_post(post) for post in raw_posts]
    
    def get_post(self, post_id: int) -> Post | None:
        self._check_client()

        # Post ID must be an int, so we convert it or raise an error if the conversion fails.
        post_id = assert_conversion(post_id, int, "post_id")
        
        raw_post = self.__client.get_post(post_id)

        return Rule34Finder.to_post(raw_post) if type(raw_post) == r34Post else None
    
    def search_comments(self, *, post_id=None, limit=None, page=None) -> list[Comment]:
        self._check_client()
        
        raw_comments = None

        if post_id == None:
            raw_comments = self._get_all_comments(page, limit)
            return [Rule34Finder.to_comment(comment) for comment in raw_comments]
        
        post_id = assert_conversion(post_id, int, "post_id")

        raw_comments = self.__client.get_comments(post_id)
        return [Rule34Finder.to_comment(post) for post in raw_comments]
    
    def get_comment(self, comment_id: int) -> Comment | None:
        self._check_client()

        comment_id = assert_conversion(comment_id, int, "comment_id")

        return None
    
    def get_notes(self, post_id: int) -> list[Note]:
        self._check_client()

        post_id = assert_conversion(post_id, int, "post_id")

        raw_notes = self.__client.get_notes(post_id)

        return [Rule34Finder.to_note(note) for note in raw_notes]
    
    def post_get_parent(self, post: Post) -> Post | None:
        if (not post.parent_id):
            return None
        
        parent_post = self.get_post(post.parent_id)

        return post._set_parent(parent_post)
    
    def post_get_children(self, post: Post) -> list[Post]:
        children_posts = self.search_posts(f"parent:{post.post_id}")

        if (not children_posts):
            return []
        
        for index, child_post in enumerate(children_posts):
            if child_post.post_id == post.post_id:
                children_posts.pop(index)
                break

        return post._set_children(children_posts)


    def _check_client(self):
        if (not isinstance(self.__client, Rule34API)):
            raise RuntimeError(f"Client must be of type Rule34API, not {type(self.__client)}")

    @property
    def configuration(self):
        return self.__config


"""
    Note.py

    Nokutoka Momiji

    Contains the Note class which represents a note associated with a post. 
    The Note class wraps information such as the note's ID, creation timestamp, coordinates, dimensions, 
    body content, source, and post ID. 
    
    It provides methods for converting the note's body to Markdown format and accessing its properties.

    This script relies on built-in Python modules such as json, datetime, and calendar for handling datetimes.
    It also utilizes the markdownify library for converting HTML-like note bodies to Markdown format.
    
"""

import json
from datetime import datetime
from calendar import timegm
from reprlib import Repr

import markdownify

class Note():
    def __init__(self, *, note_id: int, created_at: datetime, x: int, y: int, width: int, height: int, body: str, source: str, post_id: int):
        """Creates a note object from the given data.

        Args:
            note_id (int): The ID of the note.
            created_at (datetime): A datetime object containing the date and time when the note was created.
            x (int): The X coordinate of the note.
            y (int): The Y coordinate of the note.
            width (int): The width of the note.
            height (int): The height of the note.
            body (str): The text of the note. May include html tags or markdown content.
            source (str): A string identifying where the note came from. (Ex: "danbooru", "gelbooru", etc...)
            post_id (int): The ID of the post where the note is from.
        """
        self.__note_data = {}

        self.__note_data["note_id"] = note_id
        self.__note_data["created_at"] = timegm(created_at.utctimetuple())
        self.__note_data["x"] = x
        self.__note_data["y"] = y
        self.__note_data["width"] = width
        self.__note_data["height"] = height
        self.__note_data["body"] = body
        self.__note_data["source"] = source
        self.__note_data["post_id"] = post_id

    def __repr__(self) -> str:
        rep = Repr()
        pairs = ", ".join([f"{key} = {rep.repr(value)}" for key, value in self.__note_data.items()]).encode("unicode_escape")
        return f"<Note({pairs})>"

    def __str__(self) -> str:
        return json.dumps(self.__note_data)
    
    def __getitem__(self, key):
        return self.__note_data[key]
    
    def __iter__(self):
        return self.__note_data.copy().__iter__()
    
    def body_to_markdown(self) -> str:
        """Returns the contents of the note as Markdown.

        Returns:
            ``str``: String containing the text of the note in Markdown format.
        """
        return markdownify.MarkdownConverter().convert(self.body)

    @property
    def note_id(self) -> int:
        """The ID of the note."""
        return self.__note_data["note_id"]
    
    @property
    def created_at(self) -> int:
        """The timestamp when the note was created."""
        return self.__note_data["created_at"]
    
    @property
    def x(self) -> int:
        """The note's X coordinate relative to the image."""
        return self.__note_data["x"]
    
    @property
    def y(self) -> int:
        """The note's Y coordinate relative to the image."""
        return self.__note_data["y"]
    
    @property
    def width(self) -> int:
        """The width of the box containing the note."""
        return self.__note_data["width"]
    
    @property
    def height(self) -> int:
        """The height of the box containing the note."""
        return self.__note_data["height"]
    
    @property
    def body(self) -> int:
        """The text content of the note. 
        
        May contain HTML tags. Use ``body_to_markdown`` to get the body in Markdown format.
        
        """
        return self.__note_data["body"]
    
    @property
    def source(self) -> str:
        """A string representing where the note came from. (Ex: "danbooru", "gelbooru", etc...)"""
        return self.__note_data["source"]
    
    @property
    def post_id(self) -> int:
        """The ID of the post where the note is located."""
        return self.__note_data["post_id"]


from .Subfinder import *
from .DanbooruFinder import DanbooruFinder
from .Rule34Finder import Rule34Finder
from .KonachanFinder import KonachanFinder
from .YandereFinder import YandereFinder
from .GelbooruFinder import GelbooruFinder
from .PixivFinder import PixivFinder
from .ManganatoFinder import ManganatoFinder
from .NHentaiFinder import NHentaiFinder

builtin_finders = [
    DanbooruFinder,
    Rule34Finder,
    KonachanFinder,
    YandereFinder,
    GelbooruFinder,
    PixivFinder,
    ManganatoFinder
]


from enma import Enma, DefaultAvailableSources, SearchResult, Manga
from enma.domain.entities.manga import Chapter

from nokufind import Post, Comment, Note
from nokufind.Subfinder import ISubfinder, SubfinderConfiguration
from nokufind.Utils import assert_conversion, get, log

class ManganatoFinder(ISubfinder):
    @staticmethod
    def to_post(post_data: Manga | Chapter) -> Post:
        if (type(post_data) == Manga):
            return Post(
                post_id = post_data.id,
                tags = [genre.name for genre in post_data.genres],
                sources = [chapter.link.link for chapter in post_data.chapters],
                images = [post_data.cover.uri],
                authors = [author.name for author in post_data.authors],
                source = "manganato",
                preview = post_data.thumbnail.uri,
                md5 = None,
                rating = None,
                parent_id = None,
                dimensions = (post_data.cover.width, post_data.cover.height),
                poster = "",
                poster_id = None,
                name = post_data.title.english
            )
        
        return Post(
            post_id = post_data.id,
            tags = post_data.tags,
            sources = post_data.sources,
            images = [image.uri for image in post_data.pages],
            authors = post_data.authors,
            source = "manganato",
            preview = post_data.pages[0].uri,
            md5 = None,
            rating = None,
            parent_id = f"manga-{post_data.pages[0].uri.split('/')[-3]}",
            dimensions = [(page.height, page.width) for page in post_data.pages],
            poster = "",
            poster_id = None,
            name = None
        )
    
    @staticmethod
    def to_comment(comment_data) -> Comment:
        raise RuntimeError("Manganato has no comments.")
    
    @staticmethod
    def to_note(note_data) -> Note:
        raise RuntimeError("Manganato has no notes.")
    
    def __init__(self) -> None:
        source = "manganato"

        self.__client = Enma[DefaultAvailableSources]()
        self.__client.source_manager.set_source(source)

        self.__config = SubfinderConfiguration()
        self.__config._set_property("source", source)
        self.__config.set_header("Referer", "https://chapmanganato.com/")

        self.__name  = f"nokufind.Subfinder.{self.__class__.__name__}"

    def _get_all_posts(self, tags: str, limit: int = 100, page: int | None = 1):
        current_page = page if type(page) == int else 1
        limit = assert_conversion(limit, int, "limit")
        previous_size = 20
        posts = []
        oversized = True

        while (previous_size == 20):
            try:
                raw_posts = self.__client.search(tags, current_page).results
            except Exception as e:
                log(f"> [{self.__name}]: {e}")
                return posts
            
            previous_size = len(raw_posts)

            posts += raw_posts

            if (len(posts) >= limit):
                oversized = True
                break

            current_page += 1

        if (oversized):
            posts = posts[:limit]

        return posts

    def search_posts(self, tags: str | list[str], *, limit: int = 100, page: int | None = None) -> list[Post]:
        if (type(tags) == list):
            tags = " ".join(tags)

        raw_posts =  self._get_all_posts(tags, limit, page)
        posts = []
        
        for raw_post in raw_posts:
            try:
                post = self.get_post(raw_post.id)
            except:
                post = None

            if (post == None):
                continue
            posts.append(post)

        return posts

    def get_post(self, post_id: str) -> Post | None:
        log(f"> [{self.__name}]: Fetching post \"{post_id}\".")

        try:
            post = self.__client.get(post_id, with_symbolic_links = True)
        except:
            post = None

        return ManganatoFinder.to_post(post)._set_headers(self.configuration.headers) if post != None else None

    def search_comments(self, *, post_id=None, limit=None, page=None) -> list[Comment]:
        return []
    
    def get_comment(self, comment_id: int) -> Comment | None:
        return None
    
    def get_notes(self, post_id: int) -> list[Note]:
        return []
    
    def post_get_parent(self, post: Post) -> Post | None:
        if (type(post) != Post):
            raise TypeError(f"Post must be a Post object, not {type(post)}.")
        
        try:
            raw_post = self.__client.get(post.parent_id, with_symbolic_links = True)
        except:
            return None
        return ManganatoFinder.to_post(raw_post)._set_headers(self.configuration.headers) if raw_post != None else None
    
    def post_get_children(self, post: Post) -> list[Post]:
        if (type(post) != Post):
            raise TypeError(f"Post must be a Post object, not {type(post)}.")
        
        try:
            manga = self.__client.get(post.post_id)
        except:
            manga = None

        if (manga == None):
            return []
        
        for index, chapter in enumerate(manga.chapters):
            chapter.tags = manga.genres
            chapter.sources = post.sources[index]
            chapter.authors = post.authors
        
        return [ManganatoFinder.to_post(chapter)._set_headers(self.configuration.headers) for chapter in manga.chapters]

    def _headers(self):
        return self.configuration.headers

    @property
    def configuration(self):
        return self.__config


from setuptools import find_packages, setup

setup(
    name = "nokufind",
    packages = find_packages(),
    version = "1.0.0",
    description = "A library that allows you to find posts from multiple Boorus and sources.",
    author = "Nokutoka Momiji (@NokutokaMomiji)",
    install_requires = [
        "enma==2.3.0",
        "markdownify==0.11.6",
        "pixivpy3==3.7.4",
        "Pybooru==4.2.2",
        "pygelbooru==0.5.0",
        "rule34Py==1.4.11",
        "tqdm==4.66.2",
        "appdirs==1.4.4",
        "selenium==4.18.1",
        "httpx==0.27.0",
        "aiometer==0.5.0",
        "timeloop==1.0.2",
        "Requests==2.31.0"
    ],
    setup_requires = ["pytest-runner"],
    tests_require = ["pytest==4.4.1"],
    test_suite = "tests"
)


"""
    Rule34API.py

    Nokutoka Momiji

    Contains a custom Rule34API class for the Rule34Finder.

    This custom class extends the functionality of the rule34Py library, due to the sheer amount of missing 
    functionality in the original library. 

    This class allows you to fetch and parse more data from the Rule34 API, such as artists' names, notes, and comments.

    This script also includes utility functions for making requests and parsing some Rule34 XML data.

    This is mostly meant for internal use and probably requires a lot more work.
"""


import requests
import xml.etree.ElementTree as ET
from html.parser import HTMLParser

from rule34Py import rule34Py
from rule34Py import Post as r34Post
from rule34Py.post_comment import PostComment as r34Comment
from rule34Py.api_urls import API_URLS, __base_url__
from rule34Py.__vars__ import __headers__, __version__

from nokufind.Utils import USER_AGENT

POST_URL = "https://rule34.xxx/index.php?page=post&s=view&id="
NOTE_URL = "https://rule34.xxx/index.php?page=history&type=page_notes&id="
COMMENTS_URL = "https://api.rule34.xxx/index.php?page=dapi&s=comment&q=index&pid="

def _make_request(post_id: int, url: str = POST_URL):
    return requests.get(f"{url}{post_id}", headers = {"User-Agent": USER_AGENT}, cookies = {"resize-original": "1"})

class Rule34Parser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.in_artist_tag = False
        self.artist_tag_count = 0
        self.in_note_box = False
        self.in_note_body = False
        self.skipped_first_row = False
        self.in_note_table_row = False
        self.table_data_count = 0
        self.artists = []
        self.note_data = []
        self.note_table_data = []
        self.current_note_table = {}
        self.parse_note_table = False

    def handle_starttag(self, tag, attrs):
        if tag == "li" and ("class", "tag-type-artist tag") in attrs:
            self.in_artist_tag = True
        elif tag == "div" and ("class", "note-box") in attrs:
            self.in_note_box = True

            note_box_attrs = dict(attrs)
            
            self.current_note_data = {
                "width": round(float(note_box_attrs.get("style", "").split("width:")[1].split("px")[0].strip())),
                "height": round(float(note_box_attrs.get("style", "").split("height:")[1].split("px")[0].strip())),
                "x": round(float(note_box_attrs.get("style", "").split("top:")[1].split("px")[0].strip())),
                "y": round(float(note_box_attrs.get("style", "").split("left:")[1].split("px")[0].strip()))
            }
        elif tag == "div" and ("class", "note-body") in attrs:
            self.in_note_body = True
        elif tag == "tr":
            self.in_note_table_row = True

    def handle_endtag(self, tag):
        if tag == "li" and self.in_artist_tag:
            self.in_artist_tag = False
        elif tag == "div" and self.in_note_box:
            self.in_note_box = False
            self.note_data.append(self.current_note_data)
        elif tag == "div" and self.in_note_body:
            self.in_note_body = False
        elif tag == "tr":
            self.in_note_table_row = False
            if (self.skipped_first_row):
                self.note_table_data.append(self.current_note_table)
                self.current_note_table = {}
            self.skipped_first_row = True
            

    def handle_data(self, data):
        if self.in_artist_tag:
            # Skip weird empty string data.
            if not data.strip():
                return
            
            self.artist_tag_count += 1
            
            # Since there is no easy way to discern valid tag data from other stuff,
            # I just keep a counter that grabs each second element out of the three (?, tag, tag_count)
            if (self.artist_tag_count == 2):
                self.artists.append(data.strip())
                return
            
            # We reset after the third element since that is the last one in a single tag.
            if (self.artist_tag_count == 3):
                self.artist_tag_count = 0
        elif self.in_note_body:
            self.current_note_data["body"] = data.strip()
        elif self.in_note_table_row and self.skipped_first_row and self.parse_note_table:
            if not data.strip():
                return
            self.table_data_count += 1

            if (self.table_data_count == 2):
                self.current_note_table["id"] = int(data.strip())
                return
            elif (self.table_data_count == 3):
                self.current_note_table["body"] = data.strip()
                return
            elif(self.table_data_count == 5):
                self.current_note_table["created_at"] = data.strip()
                return
            
            if (self.table_data_count == 6):
                self.table_data_count = 0
            

    def get_artists(self):
        return [artist.replace(" ", "_") for artist in self.artists]
    
    def get_artists_names(self):
        return self.artists.copy()

    def get_note_data(self):
        return self.note_data.copy()
    
    def get_note_table_data(self):
        return self.note_table_data.copy()
    
    def clear_data(self):
        self.artists.clear()
        self.note_data.clear()
        self.note_table_data.clear()
        self.skipped_first_row = False
        self.parse_note_table = False

def _parse_comment_data(xml_string):
    root = ET.fromstring(xml_string)
    
    result = {
        root.tag: []
    }

    for comment_elem in root.findall('comment'):
        comment_dict = {
            'created_at': comment_elem.get('created_at'),
            'post_id': comment_elem.get('post_id'),
            'body': comment_elem.get('body'),
            'creator': comment_elem.get('creator'),
            'id': comment_elem.get('id'),
            'creator_id': comment_elem.get('creator_id')
        }
        result[root.tag].append(comment_dict)

    return result

class Rule34API(rule34Py):
    """
        Custom Rule34API class that derives from rule34Py's rule34Py class.
        
        Mainly created for adding extra data to the Post object as well as getting some data that
        cannot be obtained using the original rule34Py class, such as Notes.
    """

    # Trust me when I say that the data missing from rule34Py really infuriates me.
    # Like... the data is already there in the Rule 34 API's response data.
    # Why not... grab it? Guess we'll never know.

    def __init__(self):
        super().__init__()
        self.__parser = Rule34Parser()

    def search(self, tags: list, page_id: int = None, limit: int = 1000, deleted: bool = False, ignore_max_limit: bool = False):
        # Check if "limit" is in between 1 and 1000
        if not ignore_max_limit and limit > 1000 or limit <= 0:
            raise Exception("invalid value for \"limit\"\n  value must be between 1 and 1000\n  see for more info:\n  https://github.com/b3yc0d3/rule34Py/blob/master/DOC/usage.md#search")
            return

        params = [
            ["TAGS", "+".join(tags)],
            ["LIMIT", str(limit)],
        ]
        url = API_URLS.SEARCH.value
        # Add "page_id"
        if page_id != None:
            url += f"&pid={{PAGE_ID}}"
            params.append(["PAGE_ID", str(page_id)])

        
        if deleted:
            raise Exception("To include deleted images is not Implemented yet!")
            #url += "&deleted=show"

        formatted_url = self._parseUrlParams(url, params)
        response = requests.get(formatted_url, headers=__headers__)
        
        res_status = response.status_code
        res_len = len(response.content)
        ret_posts = []

        # checking if status code is not 200
        # (it's useless currently, becouse rule34.xxx returns always 200 OK regardless of an error)
        # and checking if content lenths is 0 or smaller
        # (curetly the only way to check for a error response)
        if res_status != 200 or res_len <= 0:
            return ret_posts

        for post in response.json():
            r34_post = r34Post.from_json(post)
            r34_post.content = post["file_url"]
            r34_post.parent_id = post["parent_id"] if post["parent_id"] else None
            r34_post.source = post["source"]
            ret_posts.append(r34_post)

        return ret_posts
    
    def get_comments(self, post_id: int) -> list:
        params = [
            ["POST_ID", str(post_id)]
        ]
        formatted_url = self._parseUrlParams(API_URLS.COMMENTS, params) # Replacing placeholders
        response = requests.get(formatted_url, headers=__headers__)

        res_status = response.status_code
        res_len = len(response.content)
        ret_comments = []

        if res_status != 200 or res_len <= 0:
            return ret_comments

        res_xml_base = _parse_comment_data(response.content.decode("utf-8"))
        res_xml = res_xml_base["comments"]

        # loop through all comments
        for comment in res_xml:
            r34_comment = r34Comment(comment["id"], comment["creator_id"], comment["body"], comment["post_id"], comment["created_at"])
            r34_comment.creator = comment["creator"]
            ret_comments.append(r34_comment)


        return ret_comments
    
    def search_comments(self, page_id: int = 1):
        request = _make_request(page_id, COMMENTS_URL)

        if (request.status_code >= 400):
            return []
        
        comments = _parse_comment_data(request.text)["comments"]
        ret_comments = []

        for comment in comments:
            r34_comment = r34Comment(comment["id"], comment["creator_id"], comment["body"], comment["post_id"], comment["created_at"])
            r34_comment.creator = comment["creator"]
            ret_comments.append(r34_comment)

        return ret_comments
    
    def get_artists(self, post_id: int) -> list[str]:
        request = _make_request(post_id)

        if (request >= 400):
            return []
        
        self.__parser.clear_data()
        self.__parser.feed(request.text)

        return self.__parser.get_artists()
    
    def get_notes(self, post_id: int) -> list[dict]:
        request = _make_request(post_id)

        if (request.status_code >= 400):
            return []
        
        self.__parser.clear_data()
        self.__parser.feed(request.text)

        note_data = self.__parser.get_note_data()
        
        for note in note_data:
            note["post_id"] = post_id
            note["id"] = None
            note["created_at"] = None

        num_of_notes = len(note_data)

        table_request = _make_request(post_id, url = NOTE_URL)
        
        if (table_request.status_code >= 400):
            return note_data
        
        self.__parser.clear_data()
        self.__parser.parse_note_table = True
        self.__parser.feed(table_request.text)

        for index, table_data in enumerate(self.__parser.get_note_table_data()):
            if (index >= num_of_notes):
                return note_data

            note = note_data[index]
            note["id"] = table_data["id"]
            note["created_at"] = table_data["created_at"]

        return note_data


from datetime import datetime, timedelta

from pixivpy3 import AppPixivAPI
from pixivpy3.utils import JsonDict
from timeloop import Timeloop

from nokufind import Post, Comment, Note
from nokufind.Subfinder import ISubfinder, SubfinderConfiguration
from nokufind.Utils import PIXIV_REFERER, get, make_request, assert_conversion
from nokufind.Utils.PixivAuth import login, refresh

PIXIV_AUTH_MESSAGE = """
"""

timeloop = Timeloop()

class PixivFinder(ISubfinder):
    @staticmethod
    def to_post(post_data: JsonDict) -> Post:
        pages = None

        if (post_data["page_count"] > 1):
            pages = [page["image_urls"]["original"] for page in post_data["meta_pages"]]
        else:
            pages = [post_data["meta_single_page"]["original_image_url"]]

        return Post(
            post_id = post_data["id"],
            tags = [tag["name"] for tag in post_data["tags"]],
            sources = [],
            images = pages,
            authors = [post_data["user"]["name"]],
            source = "pixiv",
            preview = post_data["image_urls"]["large"],
            md5 = None,
            rating = None,
            parent_id = None,
            dimensions = [(post_data["width"], post_data["height"])] * len(pages),
            poster = post_data["user"]["name"],
            poster_id = post_data["user"]["id"],
            name = None
        )
    
    @staticmethod
    def to_comment(comment_data) -> Comment:
        return Comment(
            comment_id = comment_data["id"],
            post_id = comment_data["post_id"],
            creator_id = comment_data["user"]["id"],
            creator = comment_data["user"]["name"],
            body = comment_data["comment"],
            source = "pixiv",
            created_at = datetime.strptime(comment_data["date"], "%Y-%m-%dT%H:%M:%S%z")
        )
    
    @staticmethod
    def to_note(note_data) -> Note:
        raise RuntimeError("Pixiv has no notes.")

    def __init__(self, refresh_key: str | None = None) -> None:
        self.__client = AppPixivAPI()

        if (not refresh_key):
            refresh_key = login()
            timeloop.start()

        self.__client.auth(refresh_token = refresh_key)

        self.__config = SubfinderConfiguration()
        self.__config._set_property("api_key", refresh_key)

    def search_posts(self, tags: str | list[str], *, limit: int = 100, page: int | None = None) -> list[Post]:
        self._check_client()

        if (type(tags) == list):
            tags = " ".join(tags)

        return self._get_all_posts(tags, limit, page)
    
    def get_post(self, post_id: int) -> Post | None:
        self._check_client()

        raw_post = self.__client.illust_detail(post_id)

        return PixivFinder.to_post(raw_post) if not "error" in raw_post else None
    
    def search_comments(self, *, post_id=None, limit=None, page=None) -> list[Comment]:
        comments = []
        
        if (post_id == None):
            offset = (page - 1) * 30 if page != None else 0
            while len(comments) < limit:
                posts = self.__client.illust_ranking(offset=offset)
                for post in posts.illusts:
                    raw_comment = self.__client.illust_comments(post["id"])
                    if len(raw_comment["comments"]) == 0:
                        continue

                    for comment in raw_comment["comments"]:
                        comment["post_id"] = post["id"]
                        comments.append(PixivFinder.to_comment(comment))
                offset += 30

            return comments[:limit]

        raw_comments = self.__client.illust_comments(post_id)
        if "error" in raw_comments:
            return None
        
        for comment in raw_comments["comments"]:
            comment["post_id"] = post_id

        return [PixivFinder.to_comment(comment) for comment in raw_comments]

    def get_comment(self, comment_id: int) -> Comment | None:
        self._check_client()

        return None
    
    def get_notes(self, post_id: int) -> list[Note]:
        return []
    
    def post_get_parent(self, post: Post) -> Post | None:
        return None
    
    def post_get_children(self, post: Post) -> list[Post]:
        return []
    
    # Non-standard functions
    def get_recommended_posts(self, limit: int = 100, page: int | None = None):
        self._check_client()
        
        posts = []
        limit = assert_conversion(limit, int, "limit")
        current_page = page if page != None else 0
        previous_size = 30
        oversized = False

        while previous_size == 30:
            raw_posts = self.__client.illust_recommended(offset = 30 * current_page).illusts
            previous_size = len(raw_posts)

            posts += [PixivFinder.to_post(post) for post in raw_posts]

            if (len(posts) > limit):
                oversized = True
                break
        
        if (oversized):
            return posts[:limit]
        
        return posts
    
    def get_recommended_user_posts(self, limit: int = 100, page: int | None = None):
        self._check_client()
        
        posts = []
        limit = assert_conversion(limit, int, "limit")
        current_page = page if page != None else 0
        previous_size = 30
        oversized = False

        while previous_size == 30:
            raw_posts = self.__client.user_recommended(offset = 30 * current_page).user_previews
            previous_size = len(raw_posts)

            posts += [PixivFinder.to_post(post) for post in raw_posts["illusts"]]

            if (len(posts) > limit):
                oversized = True
                break
        
        if (oversized):
            return posts[:limit]
        
        return posts

    def get_trending_tags(self):
        raise NotImplementedError()
        self.__client.trending_tags_illust()
    

    def on_config_change(self, key: str, value, is_cookie: bool, is_header: bool):
        if (key == "api_key" and value):
            self.__client.auth(refresh_token = value)

    def _get_all_posts(self, tags: str, limit: int, page: int | None = None) -> list[Post]:
        limit = assert_conversion(limit, int, "limit")
        current_page = page if page != None else 0
        previous_length = 30
        posts = []
        overflowed = False

        while (previous_length == 30):
            raw_posts = self.__client.search_illust(tags, offset = 30 * current_page).illusts
            previous_length = len(raw_posts)

            posts += [PixivFinder.to_post(post) for post in raw_posts]

            if (len(posts) > limit):
                overflowed = True
                break

            current_page += 1

        if (overflowed):
            return posts[:limit]
        
        return posts

    def _has_auth(self):
        return (self.__client.access_token != None)
    
    @timeloop.job(interval = timedelta(seconds=3500))
    def _refresh_auth(self):
        self.__client.auth(refresh_token = refresh(self.__config.get_config("api_key")))
    
    def _check_client(self):
        if (not isinstance(self.__client, AppPixivAPI)):
            raise RuntimeError(f"Client must be of type AppPixivAPI, not {type(self.__client)}.")

        if (not self._has_auth()):
            self.__config.set_config("api_key", login())


    @property
    def configuration(self):
        return self.__config


import json
from datetime import datetime
from calendar import timegm
from reprlib import Repr

import markdownify

class Comment():
    """A wrapper around comment data from multiple sources."""

    def __init__(self, *, comment_id: int, post_id: int, creator_id: int, creator: str, body: str, source: str, created_at: datetime):
        """Creates a new Comment object.

        If you are using any of the built-in Finders, you can use the to_post() static methods to automatically generate a Comment object
        from the data object itself.

        Args:
            comment_id (``int``): The ID of the comment.
            post_id (``int``): The ID of the post where the comment is located.
            creator_id (``int``): The ID of the creator of the comment.
            creator (``str``): The name of the creator of the comment.
            body (``str``): The content of the comment.
            source (``str``): A string representing the source from where the comment data came from. (Ex: "danbooru", "konachan", etc...)
            created_at (``datetime``): A datetime object containing the date information. This will be transformed into a UTC timestamp.
        """
        self.__comment_data = {}

        self.__comment_data["comment_id"] = comment_id
        self.__comment_data["post_id"] = post_id
        self.__comment_data["creator_id"] = creator_id
        self.__comment_data["creator"] = creator
        self.__comment_data["body"] = body
        self.__comment_data["source"] = source
        self.__comment_data["created_at"] = timegm(created_at.utctimetuple())

    def __repr__(self):
        rep = Repr()
        content = ", ".join([f"{key}={rep.repr(value)}" for key, value in self.__comment_data.items()])
        return f"<Comment({content})>"

    def __str__(self):
        return json.dumps(self.__comment_data)
    
    def __getitem__(self, key):
        return self.__comment_data[key]
    
    def __iter__(self):
        return self.__comment_data.copy().__iter__()
    
    def body_to_markdown(self):
        return markdownify.MarkdownConverter().convert(self.body)

    @property
    def comment_id(self) -> int:
        """The ID of the comment.

        Returns:
            int: The ID of the comment.
        """
        return self.__comment_data["comment_id"]
    
    @property
    def post_id(self) -> int:
        """The ID of the post where the comment is located.

        Returns:
            int: The ID of the post.
        """
        return self.__comment_data["post_id"]
    
    @property
    def creator_id(self) -> int:
        """The ID of the user who made the comment.

        Returns:
            int: The user ID.
        """
        return self.__comment_data["creator_id"]
    
    @property
    def creator(self) -> str:
        """The name of the user who made the comment.

        Returns:
            str: The username.
        """
        return self.__comment_data["creator"]
    
    @property
    def body(self) -> str:
        """The text content of the comment.

        Use ``body_to_markdown`` to get the body in Markdown format.

        Returns:
            str: The text of the comment.
        """
        return self.__comment_data["body"]
    
    @property
    def source(self) -> str:
        """A string representing the source where the comment came from. (Ex: "danbooru", "pixiv", etc...)"""
        return self.__comment_data["source"]
    
    @property
    def created_at(self) -> int:
        """The timestamp of when the comment was created."""
        return self.__comment_data["created_at"]